# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0
#
# The EnergyDiff model is made available via Nan Lin and Pedro P. Vergara
# from the Delft University of Technology. Nan Lin and Pedro P. Vergara are
# funded via the ALIGN4Energy Project (with project number NWA.1389.20.251) of
# the research programme NWA ORC 2020 which is (partly) financed by the Dutch
# Research Council (NWO), The Netherland.
"""
Name: diffusion class
Author: Nan Lin (sentient-codebot)
Date: Nov 2024

"""
from functools import partial
from typing import Callable, Iterator

import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from einops import reduce
from ema_pytorch import EMA
from torch import Tensor, nn
from tqdm import tqdm

from opensynth.data_modules.lcl_data_module import TrainingData
from opensynth.models.energydiff._diffusion_base import (
    BetaScheduleType,
    DiffusionBase,
    LossType,
    ModelMeanType,
    ModelPrediction,
    ModelVarianceType,
    extract,
    get_beta_schedule,
    get_loss_weight,
)
from opensynth.models.energydiff.model import DenoisingTransformer
from opensynth.models.energydiff.sampler import DPMSolverSampler


class GaussianDiffusion1D(nn.Module, DiffusionBase):
    def __init__(
        self,
        base_model: DenoisingTransformer,  # denoise model
        num_timestep: int = 1000,
        model_mean_type: ModelMeanType = ModelMeanType.NOISE,
        model_variance_type: ModelVarianceType = ModelVarianceType.FIXED_SMALL,
        loss_type: LossType = LossType.MSE,
        beta_schedule_type: BetaScheduleType = BetaScheduleType.COSINE,
    ):
        super().__init__()
        self._model = base_model
        self.dim_in = base_model.dim_in
        self.num_timestep = num_timestep
        self._model_mean_type = model_mean_type
        self._model_variance_type = model_variance_type
        self.loss_type = loss_type
        self.beta_schedule_type = beta_schedule_type
        self.dpm_sampler = None

        # Check arguments
        if not isinstance(model_mean_type, ModelMeanType):
            raise ValueError(
                "objective must be ModelMeanType.X_START, \
                ModelMeanType.NOISE, ModelMeanType.V"
            )
        if not isinstance(model_variance_type, ModelVarianceType):
            raise ValueError("model variance type must be ModelVarianceType")
        if not isinstance(beta_schedule_type, BetaScheduleType):
            raise ValueError(
                "type_beta_schedule must be BetaScheduleType.\
                linear or BetaScheduleType.cosine"
            )
        if (
            not (self.model_variance_type == ModelVarianceType.LEARNED_RANGE)
            == self.model.learn_variance
        ):
            raise ValueError(
                "denoising_model.learn_variance must be \
                consistent with diffusion_model_variance_type"
            )

        if self.model_variance_type == ModelVarianceType.LEARNED_RANGE:
            raise NotImplementedError("Learned range is not implemented yet.")
        if self.loss_type.is_vb():
            raise NotImplementedError("KL loss is not implemented yet.")

        # calculate beta_schedule
        beta_schedule = get_beta_schedule(beta_schedule_type, num_timestep)

        # other diffusion q(x_t | x_{t-1}) needed coef
        alpha = 1.0 - beta_schedule
        alpha_cumprod = torch.cumprod(
            alpha, dim=0
        )  # shape: (num_timestep,) cumulative product
        alpha_cumprod_prev = F.pad(
            alpha_cumprod[:-1], (1, 0), value=1.0
        )  # replace beginning with 1.

        self.beta_schedule = beta_schedule
        self._alpha_cumprod = alpha_cumprod
        self.alpha_cumprod_prev = alpha_cumprod_prev

        self.sqrt_alpha_cumprod = torch.sqrt(alpha_cumprod)
        self.sqrt_one_minus_alpha_cumprod = torch.sqrt(1.0 - alpha_cumprod)
        self.log_one_minus_alpha_cumprod = torch.log(1.0 - alpha_cumprod)
        self.sqrt_recip_alpha_cumprod = torch.rsqrt(alpha_cumprod)
        self.sqrt_recipm1_alpha_cumprod = torch.sqrt(1.0 / alpha_cumprod - 1.0)

        # coef for posterior q(x_{t-1} | x_t, x_0)
        posterior_variance = (
            beta_schedule * (1.0 - alpha_cumprod_prev) / (1.0 - alpha_cumprod)
        )
        self.posterior_variance = posterior_variance
        self.posterior_log_variance_clipped = torch.cat(
            [posterior_variance[1:2], posterior_variance[1:]], dim=0
        ).log()
        # clipped because variance is 0 at the beginning
        self.posterior_mean_coef1 = (
            beta_schedule
            * torch.sqrt(alpha_cumprod_prev)
            / (1.0 - alpha_cumprod)
        )
        self.posterior_mean_coef2 = (
            (1.0 - alpha_cumprod_prev)
            * torch.sqrt(alpha)
            / (1.0 - alpha_cumprod)
        )

        # calculate loss weights
        loss_weight = get_loss_weight(alpha_cumprod, model_mean_type)

        # convert to float32 for gpu
        self.beta_schedule = self.beta_schedule.to(torch.float32)
        self._alpha_cumprod = self._alpha_cumprod.to(torch.float32)
        self.alpha_cumprod_prev = self.alpha_cumprod_prev.to(torch.float32)
        self.sqrt_alpha_cumprod = self.sqrt_alpha_cumprod.to(torch.float32)
        self.sqrt_one_minus_alpha_cumprod = (
            self.sqrt_one_minus_alpha_cumprod.to(torch.float32)
        )
        self.log_one_minus_alpha_cumprod = self.log_one_minus_alpha_cumprod.to(
            torch.float32
        )
        self.sqrt_recip_alpha_cumprod = self.sqrt_recip_alpha_cumprod.to(
            torch.float32
        )
        self.sqrt_recipm1_alpha_cumprod = self.sqrt_recipm1_alpha_cumprod.to(
            torch.float32
        )
        self.posterior_variance = self.posterior_variance.to(torch.float32)
        self.posterior_log_variance_clipped = (
            self.posterior_log_variance_clipped.to(torch.float32)
        )
        self.posterior_mean_coef1 = self.posterior_mean_coef1.to(torch.float32)
        self.posterior_mean_coef2 = self.posterior_mean_coef2.to(torch.float32)
        self.loss_weight = loss_weight.to(torch.float32)

    @property
    def alpha_cumprod(self) -> torch.Tensor:
        return self._alpha_cumprod

    @property
    def model_mean_type(self) -> ModelMeanType:
        return self._model_mean_type

    @property
    def model_variance_type(self) -> ModelVarianceType:
        return self._model_variance_type

    @property
    def model(self) -> nn.Module:
        return self._model

    def predict_start_from_noise(
        self,
        x_t: Tensor,
        t: Tensor,
        noise: Tensor,
    ):
        return (
            extract(self.sqrt_recip_alpha_cumprod, t, x_t.shape) * x_t
            - extract(self.sqrt_recipm1_alpha_cumprod, t, x_t.shape) * noise
        )

    def predict_noise_from_start(self, x_t, t, x_start):
        return (
            extract(self.sqrt_recip_alpha_cumprod, t, x_t.shape) * x_t
            - x_start
        ) / extract(self.sqrt_recipm1_alpha_cumprod, t, x_t.shape)

    def predict_v(self, x_start, t, noise):
        return (
            extract(self.sqrt_alpha_cumprod, t, x_start.shape) * noise
            - extract(self.sqrt_one_minus_alpha_cumprod, t, x_start.shape)
            * x_start
        )

    def predict_start_from_v(self, x_t, t, v):
        return (
            extract(self.sqrt_alpha_cumprod, t, x_t.shape) * x_t
            - extract(self.sqrt_one_minus_alpha_cumprod, t, x_t.shape) * v
        )

    def q_mean_variance(
        self,
        x_start: Tensor,
        t: Tensor,
    ):
        "forward x_0 -> x_t, q(x_t | x_0)"
        mean = extract(self.sqrt_alpha_cumprod, t, x_start.shape) * x_start
        variance = extract(
            self.sqrt_one_minus_alpha_cumprod**2, t, x_start.shape
        )
        log_variance = extract(
            self.log_one_minus_alpha_cumprod, t, x_start.shape
        )
        return mean, variance, log_variance

    def q_posterior_mean_variance(
        self,
        x_start: Tensor,
        x_t: Tensor,
        t: Tensor,
    ):
        "posterior q(x_{t-1} | x_t, x_0)"
        posterior_mean = (
            extract(self.posterior_mean_coef1, t, x_t.shape) * x_start
            + extract(self.posterior_mean_coef2, t, x_t.shape) * x_t
        )
        posterior_variance = extract(
            self.posterior_variance, t, x_t.shape
        )  # shape: (batch, 1, 1)
        posterior_log_variance_clipped = extract(
            self.posterior_log_variance_clipped, t, x_t.shape
        )

        return (
            posterior_mean,
            posterior_variance,
            posterior_log_variance_clipped,
        )

    def q_sample(
        self,
        x_start: Tensor,
        t: Tensor,
        noise: None | Tensor = None,
    ) -> Tensor:
        "sample x_t from q(x_t | x_0)"
        noise = noise if noise is not None else torch.randn_like(x_start)
        return (
            extract(self.sqrt_alpha_cumprod, t, x_start.shape) * x_start
            + extract(self.sqrt_one_minus_alpha_cumprod, t, x_start.shape)
            * noise
        )

    def model_prediction(
        self,
        x_t: Tensor,
        t: Tensor,
        clip_x_start: bool = False,
    ) -> ModelPrediction:
        "get ModelPrediction(noise, x_start, var_factor)"
        model_output = self.model(x_t, t)
        _clip_fn: Callable = partial(torch.clamp, min=-1.0, max=1.0)
        _identity_fn: Callable = lambda x: x
        maybe_clip = _clip_fn if clip_x_start else _identity_fn

        # if learned variance range, split the mean and variance prediction
        if self.model_variance_type == ModelVarianceType.LEARNED_RANGE:
            if not model_output.shape == (
                x_t.shape[0],
                x_t.shape[1],
                2 * x_t.shape[2],
            ):
                raise ValueError(
                    "model output shape does not match expected shape."
                )

            model_output, model_var_factor = torch.split(
                model_output, x_t.shape[2], dim=-1
            )
        else:
            model_var_factor = None

        # calculate noise and x_start depending on model_mean_type
        if self.model_mean_type == ModelMeanType.NOISE:
            pred_noise = model_output
            x_start = self.predict_start_from_noise(x_t, t, pred_noise)
            x_start = maybe_clip(x_start)
            # optional, rederive noise from x_start after clipping
            pred_noise = self.predict_noise_from_start(x_t, t, x_start)
        elif self.model_mean_type == ModelMeanType.X_START:
            x_start = model_output
            x_start = maybe_clip(x_start)
            pred_noise = self.predict_noise_from_start(x_t, t, x_start)
        elif self.model_mean_type == ModelMeanType.V:
            v = model_output
            x_start = self.predict_start_from_v(x_t, t, v)
            x_start = maybe_clip(x_start)
            pred_noise = self.predict_noise_from_start(x_t, t, x_start)
        else:
            raise NotImplementedError(
                f"model_mean_type: {self.model_mean_type} not implemented."
            )

        return ModelPrediction(pred_noise, x_start, model_var_factor)

    def p_mean_variance(
        self,
        x_t: Tensor,
        t: Tensor,
        clip_denoised: bool = False,
        model_kwargs=None,
    ) -> dict:
        """Get approximate mean and var of posterior p_model(x_{t-1} | x_t).
            Also returns x_start
        :param x_t: the input x_t, float, shape: (batch, sequence, dim)
        :param t: the time step, int, shape: (batch,)
        :param clip_denoised: whether to
            clip the denoised x_start to [-1, 1]
        :param model_kwargs: additional kwargs for \
            for model_prediction function

        :return dict: {
            'model_mean': model_mean
            'model_variance': model_variance
            'model_log_variance': model_log_variance
            'pred_x_start': pred_x_start
        }
        """
        model_kwargs = model_kwargs if model_kwargs is not None else {}
        pred = self.model_prediction(x_t, t, **model_kwargs)
        pred_x_start = pred.pred_x_start
        var_factor = pred.pred_var_factor

        # get variance
        if self.model_variance_type == ModelVarianceType.LEARNED_RANGE:
            min_log = extract(
                self.posterior_log_variance_clipped, t, x_t.shape
            )
            max_log = extract(torch.log(self.beta_schedule), t, x_t.shape)
            frac = (var_factor + 1.0) / 2.0  # [-1, 1] -> [0, 1]
            model_log_variance = frac * max_log + (1.0 - frac) * min_log
            model_variance = model_log_variance.exp()
        else:
            model_variance, model_log_variance = {
                ModelVarianceType.FIXED_LARGE: (
                    torch.cat(
                        [self.posterior_variance[1:2], self.beta_schedule[1:]],
                        dim=0,
                    ),
                    torch.log(
                        torch.cat(
                            [
                                self.posterior_variance[1:2],
                                self.beta_schedule[1:],
                            ],
                            dim=0,
                        )
                    ),
                ),
                ModelVarianceType.FIXED_SMALL: (
                    self.posterior_variance,
                    self.posterior_log_variance_clipped,
                ),
            }[self.model_variance_type]
            model_log_variance = extract(model_log_variance, t, x_t.shape)
            model_variance = extract(model_variance, t, x_t.shape)

        # get mean
        if clip_denoised:
            pred_x_start = torch.clamp(pred_x_start, min=-1.0, max=1.0)

        model_mean, *_ = self.q_posterior_mean_variance(pred_x_start, x_t, t)

        if not (
            model_mean.shape
            == model_log_variance.shape
            == model_variance.shape
            == x_t.shape
        ):
            raise ValueError(
                f"inconsistent shape: {model_mean.shape}, \
                {model_log_variance.shape}, {model_variance.shape}, \
                    {x_t.shape}"
            )

        return {
            "model_mean": model_mean,
            "model_variance": model_variance,
            "model_log_variance": model_log_variance,
            "pred_x_start": pred_x_start,
        }

    @torch.no_grad()
    def p_sample(
        self,
        x_t: Tensor,
        t: int,  # int, scalar, not a batched tensor.
        clip_denoised: bool = False,
        model_kwargs: None | dict = None,
        noise: None | Tensor = None,
    ) -> Tensor:
        """Apply p_model(x_{t-1} | x_t) to sample x_{t-1} from x_t.
        A inference-centric function, so `t` is an integer,
            same for every sample in the batch.
        :param x_t: x_t, shape (batch, channel, sequence)
        :param t: timestep, int
        :param clip_denoised: whether to clip the denoised x_{t-1} to [-1, 1]
        :param model_kwargs: kwargs for model_prediction
        :param noise: optional. noise to use
            for sampling, shape (batch, channel, sequence)

        :return: {
            'pred_x_prev': pred_x_prev,
            'pred_x_start': out['pred_x_start'],
        }
        """
        b, *_ = x_t.shape
        batched_time = torch.full((b,), t, device=x_t.device, dtype=torch.long)
        out_dict = self.p_mean_variance(
            x_t,
            t=batched_time,
            clip_denoised=clip_denoised,
            model_kwargs=model_kwargs,
        )
        if noise is None or noise.shape != x_t.shape:
            noise = torch.randn_like(x_t) if t > 0 else 0.0
            # if t == 0, noise is not needed
        pred_x_prev = (
            out_dict["model_mean"]
            + (0.5 * out_dict["model_log_variance"]).exp() * noise
        )

        return {
            "pred_x_prev": pred_x_prev,
            "pred_x_start": out_dict["pred_x_start"],
        }

    @torch.no_grad()
    def p_sample_loop_progressive(
        self,
        shape: torch.Size,  # 'b l d'
        noise: None | Tensor = None,
        clip_denoised: bool = False,
        model_kwargs: None | dict = None,
    ) -> Iterator[dict[str, Tensor]]:
        """Creates a generator that progressively samples from the diffusion \
            process from x_T to x_0.

        This method implements the progressive sampling loop that generates \
            samples by iteratively
        denoising from timestep T-1 to 0. At each step, it applies the \
            p_sample function to
        predict the previous timestep's sample.

        Args:
            shape (torch.Size): The shape of the tensor to sample, \
                expected in format 'b l d'
                (batch, length, dimension).
            noise (Optional[Tensor], optional): Initial noise tensor. \
                If None, random noise will be used.
                Defaults to None.
            clip_denoised (bool, optional): Whether to clip the denoised \
                values. Defaults to False.
            model_kwargs (Optional[dict], optional): Additional arguments \
                to pass to the model.
                Defaults to None.

        Yields:
            dict[str, Tensor]: Dictionary containing sampling information at \
                each timestep, including:
                - pred_x_prev: The predicted sample for the previous timestep
                - pred_noise: The predicted noise component
                - Other model-specific outputs

        Returns:
            Iterator[dict[str, Tensor]]: Generator yielding progressive \
                sampling results from T-1 to 0.
        """
        device = next(self.model.parameters()).device
        model_kwargs = model_kwargs

        x_t = noise if noise is not None else torch.randn(shape, device=device)
        # the initial x_t, t = T

        # yielding T-1 -> 0
        for t in tqdm(
            reversed(range(0, self.num_timestep)),
            desc="Sampling Loop Timestep",
            total=self.num_timestep,
        ):
            sample_out = self.p_sample(
                x_t,
                t=t,  # int
                clip_denoised=clip_denoised,
                model_kwargs=model_kwargs,
            )  # the noise parameter
            # IS NOT THE SAME AS the noise here.
            yield sample_out
            x_t = sample_out["pred_x_prev"]

    def p_sample_loop(
        self,
        shape: torch.Size,  # 'b l d'
        noise: None | Tensor = None,
        clip_denoised: bool = False,
        model_kwargs: None | dict = None,
    ) -> Tensor:
        "return the final x_0"
        for sample_out in self.p_sample_loop_progressive(
            shape, noise, clip_denoised, model_kwargs
        ):
            sample = sample_out["pred_x_prev"]

        return sample

    def _vb_terms_bpd(self, *args, **kwargs):
        raise NotImplementedError(
            "Not really useful. Do not use learned range."
        )

    def train_losses(
        self,
        x_start: Tensor,  # true x_0
        t: Tensor,  # timestep
        noise: None | Tensor = None,  # noise
        model_kwargs: None | dict = None,  # kwargs for model_prediction
    ) -> dict[str, Tensor]:
        """return batched loss. not yet averaged over batch.

        given
        - x_start: true data samples, shape (batch, channel, sequence)
        - t: randomly sampled (diffusion) timestep, shape (batch,)
        - noise: optional, noise to
            impose on x_start, shape (batch, channel, sequence)

        apply forward diffusion to get x_t;
        apply model to get approximate posterior parameters;
        calculate loss on the posterior parameters.

        :return dict: loss_terms: {
            'loss': shape (batch,)
            'mse': shape (batch,), if applicable
            'vb': vb, shape (batch,), if applicable
        }
        """
        # forward, get x_t
        noise = noise if noise is not None else torch.randn_like(x_start)
        x_t = self.q_sample(x_start, t, noise)

        loss_terms = {}
        # get model prediction
        if self.loss_type in [LossType.KL, LossType.RESCALED_KL]:
            raise NotImplementedError("KL loss is not implemented yet.")
        elif self.loss_type in [LossType.MSE, LossType.RESCALED_MSE]:
            out = self.p_mean_variance(
                x_t, t, clip_denoised=False, model_kwargs=model_kwargs
            )

            if self.model_variance_type == ModelVarianceType.LEARNED_RANGE:
                raise NotImplementedError(
                    "Learned range is not implemented yet."
                )

            # mse terms
            target = {
                ModelMeanType.NOISE: noise,
                ModelMeanType.X_START: x_start,
                ModelMeanType.V: self.predict_v(x_start, t, noise),
            }[self.model_mean_type]
            pred_noise = self.predict_noise_from_start(
                x_t, t, out["pred_x_start"]
            )
            source = {
                ModelMeanType.NOISE: pred_noise,
                ModelMeanType.X_START: out["pred_x_start"],
                ModelMeanType.V: self.predict_v(
                    out["pred_x_start"], t, pred_noise
                ),
            }[self.model_mean_type]
            if not target.shape == source.shape:
                raise ValueError("target and source shape mismatch.")

            mse_loss = reduce(
                F.mse_loss(source, target, reduction="none"),
                "b ... -> b",
                "mean",
            )  # shape (batch, )
            loss_terms["mse"] = mse_loss
            loss_terms["loss"] = mse_loss
        else:
            raise ValueError(
                "loss_type must be one of MSE, RESCALED_MSE, KL, RESCALED_KL"
            )

        return loss_terms

    def forward(
        self,
        x_start: Tensor,  # true x_0
        noise: None | Tensor = None,  # noise
    ) -> dict[str, Tensor]:
        "return scalar loss. averaged over batch already."
        B, L, D = x_start.shape
        device = x_start.device
        if not D == self.dim_in:
            raise ValueError(f"Dimension mismatch: {D} != {self.dim_in}")

        t = torch.randint(0, self.num_timestep, (B,), device=device)
        loss_terms = self.train_losses(x_start, t, noise)

        for k, v in loss_terms.items():
            loss_terms[k] = v.mean()

        return loss_terms

    # two functions we use to sample.
    def ancestral_sample(
        self,
        num_sample: int,
        batch_size: int,
        data_sequence_length: int,
        data_feature_dim: int,  # 1 for univariate, 2 for bivariate
        # cond: torch.Tensor|None=None,
        # cfg_scale: float = 1.,
    ) -> torch.Tensor:
        """sample from either given diffusion model.

        Arguments:
            - num_sample: int, number of samples to generate
            - batch_size: int, batch size for sampling
            - cond: shape [batch_size, channel, 1], condition for sampling
        """
        # process cond
        # if cond is not None:
        #     if model is not None:
        #         _device = next(model.parameters()).device
        #         cond = cond.to(_device)
        #     else:
        #         cond = cond.to(trainer.device)

        # process batch size split
        if num_sample < batch_size:
            list_batch_size = [num_sample]
        else:
            list_batch_size = [batch_size] * (num_sample // batch_size)
            if num_sample % batch_size != 0:
                list_batch_size.append(num_sample % batch_size)

        # sample
        list_sample = []
        for idx, batch_size in enumerate(list_batch_size):
            print(
                f"sampling batch {idx + 1}/{len(list_batch_size)}, \
                    batch size {batch_size}. "
            )
            sample_batch = self.p_sample_loop(
                shape=(batch_size, data_sequence_length, data_feature_dim),
                noise=None,
                clip_denoised=False,  # True requires data scaling to (-1,1)
            )

            list_sample.append(sample_batch)
        all_sample = torch.cat(list_sample, dim=0)

        return all_sample

    def _init_dpm_sampler(self):
        self.dpm_sampler = DPMSolverSampler(self)

    def dpm_solver_sample(
        self,
        total_num_sample: int,
        batch_size: int,
        step: int,
        shape: tuple[int, int],
        # conditioning: torch.Tensor|None,
        # cfg_scale: float,
        clip_denoised: bool = False,
    ) -> torch.Tensor:
        if step < 15:
            raise ValueError(
                "sampling step must be at least 15 due to\
                constraint of DPM-Solver."
            )
        if self.dpm_sampler is None:
            self._init_dpm_sampler()
        assert self.dpm_sampler is not None

        num_sample = total_num_sample
        if num_sample < batch_size:
            list_batch_size = [num_sample]
        else:
            list_batch_size = [batch_size] * (num_sample // batch_size)
            if num_sample % batch_size != 0:
                list_batch_size.append(num_sample % batch_size)

        list_sample = []
        for idx, batch_size in enumerate(list_batch_size):
            print(
                f"sampling batch {idx + 1}/{len(list_batch_size)}, \
                    batch size {batch_size}. "
            )
            sample_batch, _ = self.dpm_sampler.sample(
                S=step,
                batch_size=batch_size,
                shape=shape,
                # conditioning = conditioning,
                # cfg_scale = cfg_scale,
            )
            list_sample.append(sample_batch)

        all_sample = torch.cat(list_sample, dim=0)

        if clip_denoised:
            all_sample = torch.clamp(all_sample, -1, 1)
        return all_sample


class PLDiffusion1D(pl.LightningModule):
    "Pytorch Lightning wrapper for GaussianDiffusion1D"

    def __init__(
        self,
        # transformer model arguments
        dim_base: int = 256,
        dim_in: int = 1,
        num_attn_head: int = 4,
        num_decoder_layer: int = 6,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        learn_variance: bool = False,
        # diffusion model arguments
        num_timestep: int = 1000,
        model_mean_type: ModelMeanType = ModelMeanType.NOISE,
        model_variance_type: ModelVarianceType = ModelVarianceType.FIXED_SMALL,
        loss_type: LossType = LossType.MSE,
        beta_schedule_type: BetaScheduleType = BetaScheduleType.COSINE,
        # PL optimizer arguments
        lr: float = 1e-4,
        ema_update_every: int = 5,
        ema_decay: float = 0.9999,
        disable_init_proj: bool = False,
    ):
        super().__init__()
        self.save_hyperparameters()
        base_model = DenoisingTransformer(
            dim_base=dim_base,
            dim_in=dim_in,
            num_attn_head=num_attn_head,
            num_decoder_layer=num_decoder_layer,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            learn_variance=learn_variance,
            disable_init_proj=disable_init_proj,
        )
        self.diffusion_model = GaussianDiffusion1D(
            base_model=base_model,
            num_timestep=num_timestep,
            model_mean_type=model_mean_type,
            model_variance_type=model_variance_type,
            loss_type=loss_type,
            beta_schedule_type=beta_schedule_type,
        )
        self.ema = EMA(
            self.diffusion_model,
            beta=ema_decay,
            update_every=ema_update_every,
            include_online_model=True,
        )
        self.lr = lr
        self.ema_update_every = ema_update_every
        self.ema_decay = ema_decay

    # setup function
    def setup(self, stage: str) -> None:
        pass  # nothing to setup

    def on_before_batch_transfer(
        self,
        batch: TrainingData,
        dataloader_idx: int,
    ):
        kwh_data = batch["kwh"]  # shape: (batch, sequence)
        kwh_data = kwh_data.unsqueeze(-1)  # shape: (batch, sequence, 1)
        return kwh_data

    # training step
    def training_step(
        self,
        batch: Tensor,
        batch_idx: int,  # -> step counter
    ) -> Tensor:
        loss_terms = self.diffusion_model(x_start=batch)
        loss = loss_terms["loss"]
        mse = loss_terms["mse"].item() if "mse" in loss_terms else 0.0
        self.log(
            "train_loss",
            loss.item(),
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            sync_dist=True,
        )
        self.log(
            "train_mse_loss",
            mse,
            on_step=True,
            on_epoch=True,
            sync_dist=True,
        )

        return loss

    def on_train_batch_end(self, outputs, batch, batch_idx):
        self.ema.update()

    # validation step
    @torch.no_grad()
    def validation_step(self, batch, batch_idx):
        loss_terms = self.diffusion_model(x_start=batch)
        loss = loss_terms["loss"]
        mse = loss_terms["mse"].item() if "mse" in loss_terms else 0.0
        self.log(
            "val_loss",
            loss,
            on_step=False,
            on_epoch=True,
            sync_dist=True,
        )
        self.log(
            "val_mse_loss",
            mse,
            on_step=False,
            on_epoch=True,
            sync_dist=True,
        )

    # test step
    @torch.no_grad()
    def test_step(self, batch, batch_idx):
        raise NotImplementedError("Validation step is not implemented yet.")

    # configure optimizer
    def configure_optimizers(self) -> torch.optim.Optimizer:
        optimizer = torch.optim.AdamW(
            self.diffusion_model.parameters(), lr=self.lr
        )  # just online model
        return optimizer



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

from opensynth.models.faraday.gaussian_mixture.gmm_init import (
    initialise_gmm_params,
)
from opensynth.models.faraday.gaussian_mixture.gmm_model import (
    GaussianMixtureLightningModule,
    GaussianMixtureModel,
)

__all__ = [
    "GaussianMixtureModel",
    "GaussianMixtureLightningModule",
    "initialise_gmm_params",
]



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import ast
import random
from pathlib import Path
from typing import Optional, TypedDict

import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
from torch.utils.data import DataLoader, Dataset

RANDOM_STATE = 0
g = torch.Generator()
g.manual_seed(RANDOM_STATE)


def seed_worker(worker_id):
    worker_seed = torch.initial_seed() % 2**32
    np.random.seed(worker_seed)
    random.seed(worker_seed)


class TrainingData(TypedDict):
    kwh: torch.Tensor
    features: dict[str, torch.Tensor]


class LCLData(Dataset):
    """
    Low CarbonLondon Dataset. The dataset should return
    TrainingData(TypedDict) which contains:
    - kwh: kWh data
    - features: Dictionary of features

    To use Faraday on custom datasets, your data module
    should also return data in the same format.
    """

    def __init__(
        self,
        data_path: Path,
        stats_path: Path,
        n_samples: int,
        outlier_path: Optional[Path] = None,
    ):
        """
        Args:
            data_path (Path): Data path
            stats_path (Path): Stats path. Note when loading evaluation
            dataset, the stats path point to the stats of training data,
            rather than evaluation data to avoid data leakage!
            n_samples (int): Number of samples to load
            outlier_path (Path, optional): Path to outlier data.
            Defaults to None.
        """
        self.df = pd.read_csv(data_path)
        self.df_stats = pd.read_csv(stats_path)
        self.outlier = True if outlier_path else False

        # Parse stats
        self.feature_mean = self.df_stats["mean"].values[0]
        self.feature_std = self.df_stats["stdev"].values[0]

        # Resample Dataset
        self.n_samples = n_samples
        self.df = self.df.sample(
            self.n_samples, random_state=RANDOM_STATE
        ).reset_index(drop=True)

        # Combine with outliers:
        if self.outlier:
            self.df_outliers = pd.read_csv(outlier_path)
            self.df = pd.concat([self.df, self.df_outliers])
            self.df = self.df.sample(
                frac=1, random_state=RANDOM_STATE
            ).reset_index(drop=True)

        # Parse columns
        self.kwh = self.df["kwh"].apply(ast.literal_eval)
        self.kwh = torch.from_numpy(np.array(self.kwh.tolist())).float()
        self.month = self.df["month"]
        self.dayofweek = self.df["dayofweek"]

    def standardise(self, x: torch.Tensor) -> torch.Tensor:
        """
        Standardise kWh with mean 0 and std 1

        Args:
            x (torch.Tensor): Input kWh

        Returns:
            torch.Tensor: Standardised kWh
        """
        return (x - self.feature_mean) / self.feature_std

    def reconstruct(self, xhat: torch.Tensor) -> torch.Tensor:
        """
        Reconstruct kWh from standardised values

        Args:
            xhat (torch.Tensor): standardised kWh

        Returns:
            torch.Tensor: reconstructed kWh
        """
        return (xhat * self.feature_std) + self.feature_mean

    def __len__(self):
        return len(self.df)

    def __getitem__(self, idx):
        standardised_kwh = self.standardise(self.kwh[idx])
        features: dict[str, torch.Tensor] = {
            "month": self.month[idx],
            "dayofweek": self.dayofweek[idx],
        }
        return TrainingData(kwh=standardised_kwh, features=features)


class LCLDataModule(pl.LightningDataModule):
    """
    Low Carbon London data module
    """

    def __init__(
        self,
        data_path: Path,
        stats_path: Path,
        batch_size: int,
        n_samples: int,
        outlier_path: Optional[Path] = None,
    ):
        super().__init__()
        self.data_path = data_path
        self.stats_path = stats_path
        self.batch_size = batch_size
        self.n_samples = n_samples
        self.outlier_path = outlier_path
        self.outlier = True if outlier_path else False

    def prepare_data(self):
        pass

    def setup(self, stage=""):

        self.dataset = LCLData(
            data_path=self.data_path,
            stats_path=self.stats_path,
            n_samples=self.n_samples,
            outlier_path=self.outlier_path,
        )

        if self.outlier:
            self.outlier_dataset = LCLData(
                data_path=self.outlier_path,
                stats_path=self.stats_path,
                n_samples=100,  # Outlier size = 100
            )

    def train_dataloader(self):
        return DataLoader(
            self.dataset,
            self.batch_size,
            drop_last=True,
            shuffle=False,
            generator=g,
            worker_init_fn=seed_worker,
        )

    def outlier_dataloader(self):
        return DataLoader(
            self.outlier_dataset,
            100,
            drop_last=True,
            shuffle=False,
            generator=g,
            worker_init_fn=seed_worker,
        )

    def reconstruct_kwh(self, xhat: torch.Tensor) -> torch.Tensor:
        return self.dataset.reconstruct(xhat)



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0
#
# The EnergyDiff model is made available via Nan Lin and Pedro P. Vergara
# from the Delft University of Technology. Nan Lin and Pedro P. Vergara are
# funded via the ALIGN4Energy Project (with project number NWA.1389.20.251) of
# the research programme NWA ORC 2020 which is (partly) financed by the Dutch
# Research Council (NWO), The Netherland.
from opensynth.models.energydiff.diffusion import PLDiffusion1D

__all__ = ["PLDiffusion1D"]



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import ast
from pathlib import Path
from typing import Union

import litdata as ld
import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch

from opensynth.data_modules.lcl_data_module import TrainingData

RANDOM_STATE = 0
g = torch.Generator()
g.manual_seed(RANDOM_STATE)


class StreamData(ld.StreamingDataset):
    """
    Streaming Dataset. The dataset should return
    TrainingData(TypedDict) which contains:
    - kwh: kWh data
    - features: Dictionary of features

    To use Faraday on custom datasets, your data module
    should also return data in the same format.

    For more information on the attributes used for the StreamingDataset, see:
    https://github.com/Lightning-AI/litdata/blob/main/src/litdata/streaming/dataset.py#L43
    """

    def __init__(
        self,
        data_path: Path,
        stats_path: Path,
        storage_options: Union[dict, None] = None,
        subsample: float = 1.0,
        shuffle: bool = False,
        drop_last: bool = True,
        max_pre_download: int = 2,
        max_cache_size: str = "100GB",
    ):
        """
        Args:
            data_path (str): Path to stream data from (can be GCS URI or local
                path)
            stats_path (Path): Stats path. Note when loading evaluation
                dataset, the stats path point to the stats of training data,
                rather than evaluation data to avoid data leakage!
            storage_options (dict): Storage options metadata (e.g. GCS project)
            feature_mean (float): Mean of consumption data (kWh), used for
                standardisation
            feature_std (float): STD of consumption data (kWh), used for
                standardisation
            shuffle (bool, optional): Whether to shuffle the data
            drop_last (bool, optional): Drop last items so processes return
                same amount of data
        """
        super().__init__(
            input_dir=data_path,
            storage_options=storage_options,
            shuffle=shuffle,
            drop_last=drop_last,
            subsample=subsample,
            max_pre_download=max_pre_download,
            max_cache_size=max_cache_size,
        )
        # Parse stats
        self.df_stats = pd.read_csv(stats_path)
        self.feature_mean = self.df_stats["mean"].values[0]
        self.feature_std = self.df_stats["stdev"].values[0]

    def standardise(self, x: torch.Tensor) -> torch.Tensor:
        """
        Standardise kWh with mean 0 and std 1
        Args:
            x (torch.Tensor): Input kWh
        Returns:
            torch.Tensor: Standardised kWh
        """
        return (x - self.feature_mean) / self.feature_std

    def reconstruct(self, xhat: torch.Tensor) -> torch.Tensor:
        """
        Reconstruct kWh from standardised values
        Args:
            xhat (torch.Tensor): standardised kWh
        Returns:
            torch.Tensor: reconstructed kWh
        """
        return (xhat * self.feature_std) + self.feature_mean

    def __getitem__(self, idx):
        data = super().__getitem__(idx)

        # Parse kwh
        kwh = ast.literal_eval(data["kwh"])
        kwh = torch.from_numpy(np.array(kwh)).float()
        standardised_kwh = self.standardise(kwh)

        return TrainingData(
            kwh=standardised_kwh,
            features=data["features"],
        )


class StreamDataModule(pl.LightningDataModule):
    def __init__(
        self,
        data_path: Path,
        stats_path: Path,
        storage_options: Union[dict, None] = None,
        num_workers: int = 0,
        batch_size: int = 5000,
        subsample: float = 1.0,
        max_cache_size: str = "100GB",
        profile_batches: Union[int, None] = None,
        shuffle: bool = False,
        drop_last: bool = True,
        persistent_workers: bool = False,
        pin_memory: bool = True,
    ):
        """
        Faraday Streaming Lightning Data Module.

        For more information on the attributes used for the
        StreamingDataLoader, see:
        https://github.com/Lightning-AI/litdata/blob/main/src/litdata/streaming/dataloader.py#L494

        Args:
            data_path (str): Input streaming directory (GCS or local)
            stats_path (Path): Stats path. Note when loading evaluation
                dataset, the stats path point to the stats of training data,
                rather than evaluation data to avoid data leakage!
            storage_options (dict): Storage options metadata (e.g. GCS project)
            num_workers (int, optional): How many subprocesses to use for
                dataloading
            batch_size (int, optional): Samples per batch. Defaults to 5000
            subsample (float, optional): Float representing fraction of the
                dataset to be randomly sampled
            profile_batches (int, optional): Whether to record data loading
                profile and generate a result.json file (for debugging).
                Value is the number of batches to profile. Defaults to None.
            max_cache_size (str, optional): The maximum cache size used by
                the StreamingDataset
            shuffle (bool, optional): Whether to shuffle the data
            drop_last (bool, optional): Drop last items so processes return
                same amount of data
            persistent_workers (bool, optional): Whether to keep workers
                alive between epochs. Defaults to False
            pin_memory (bool, optional): Whether to pin memory in DataLoader.
                Defaults to True
        """
        super().__init__()
        self.data_path = data_path
        self.stats_path = stats_path
        self.storage_options = storage_options
        self.num_workers = num_workers
        self.batch_size = batch_size
        self.subsample = subsample
        self.profile_batches = profile_batches
        self.max_cache_size = max_cache_size
        self.shuffle = shuffle
        self.drop_last = drop_last
        self.persistent_workers = persistent_workers
        self.pin_memory = pin_memory

    def prepare_data(self):
        pass

    def setup(self, stage: str = ""):
        self.train_dataset = StreamData(
            data_path=self.data_path,
            stats_path=self.stats_path,
            storage_options=self.storage_options,
            shuffle=self.shuffle,
            drop_last=self.drop_last,
            subsample=self.subsample,
            max_cache_size=self.max_cache_size,
        )

    def train_dataloader(self):
        return ld.StreamingDataLoader(
            dataset=self.train_dataset,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
            persistent_workers=self.persistent_workers,
            profile_batches=self.profile_batches,
            pin_memory=self.pin_memory,
        )

    def reconstruct_kwh(self, xhat: torch.Tensor) -> torch.Tensor:
        return self.train_dataset.reconstruct(xhat)



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

from opensynth.models.faraday.model import FaradayModel
from opensynth.models.faraday.vae_model import (
    Decoder,
    Encoder,
    FaradayVAE,
    ReparametrisationModule,
)

__all__ = [
    "FaradayVAE",
    "FaradayModel",
    "Encoder",
    "Decoder",
    "ReparametrisationModule",
]



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import torch
from torchmetrics import Metric


class WeightsMetric(Metric):
    full_state_update = False

    def __init__(self, num_components: int):
        super().__init__()
        self.weights: torch.Tensor

        self.add_state(
            "weights", torch.zeros(num_components), dist_reduce_fx="mean"
        )

    def update(self, weights: torch.Tensor) -> None:
        self.weights.add_(weights)

    def compute(self) -> None:
        return self.weights


class MeansMetric(Metric):
    full_state_update = False

    def __init__(self, num_components: int, num_features: int):
        super().__init__()
        self.means: torch.Tensor
        self.add_state(
            "means",
            torch.zeros(num_components, num_features),
            dist_reduce_fx="mean",
        )

    def update(self, means: torch.Tensor) -> None:
        self.means.add_(means)

    def compute(self) -> None:
        return self.means


class PrecisionCholeskyMetric(Metric):
    full_state_update = False

    def __init__(self, num_components: int, num_features: int):
        super().__init__()
        self.precision_cholesky: torch.Tensor
        self.add_state(
            "precision_cholesky",
            torch.zeros(num_components, num_features, num_features),
            dist_reduce_fx="mean",
        )

    def update(self, precision_cholesky: torch.Tensor) -> None:
        self.precision_cholesky.add_(precision_cholesky)

    def compute(self) -> None:
        return self.precision_cholesky


class CovarianceMetric(Metric):
    full_state_update = False

    def __init__(self, num_components: int, num_features: int):
        super().__init__()
        self.covariances: torch.Tensor
        self.add_state(
            "covariances",
            torch.zeros(num_components, num_features, num_features),
            dist_reduce_fx="mean",
        )

    def update(self, covariances: torch.Tensor) -> None:
        self.covariances.add_(covariances)

    def compute(self) -> None:
        return self.covariances


class NegativeLogLikelihoodMetric(Metric):
    full_state_update = False

    def __init__(self):
        super().__init__()
        self.nll: torch.Tensor
        self.add_state(
            "nll",
            torch.zeros(1),
            dist_reduce_fx="mean",
        )

    def update(self, nll: torch.Tensor) -> None:
        self.nll.add_(nll)

    def compute(self) -> None:
        return self.nll



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0
#
# The EnergyDiff model is made available via Nan Lin and Pedro P. Vergara
# from the Delft University of Technology. Nan Lin and Pedro P. Vergara are
# funded via the ALIGN4Energy Project (with project number NWA.1389.20.251) of
# the research programme NWA ORC 2020 which is (partly) financed by the Dutch
# Research Council (NWO), The Netherland.
"""
Name: nn modules for diffusion
Author: Nan Lin (sentient-codebot)
Date: Nov 2024

"""
import math

import torch
import torch.nn.functional as F
from einops import einsum, rearrange
from einops.layers.torch import Rearrange
from torch import Tensor, nn

# model components


class SinusoidalPosEmb(nn.Module):
    """for position t, dimension i of a d-dim vector, the embedding is
    1/(10000**(i/(d/2-1)))

    dim must be even.

    Args:
        dim (int): dimension of the input

    Forward:
        pos: (batch, sequence)

    Return:
        emb: (batch, dim)
    """

    def __init__(self, dim: int):
        super().__init__()
        if dim % 2 != 0:
            raise ValueError("dimension must be even")
        self.dim = dim

    def forward(self, pos: Tensor) -> Tensor:
        """
        Args:
            pos: (batch, sequence)
        """
        device = pos.device
        half_dim = self.dim // 2
        emb = torch.exp(
            torch.arange(half_dim, device=device)
            * -math.log(10000)
            / (half_dim - 1)
        )  # shape: (dim/2,)
        emb = pos.unsqueeze(-1) * emb.unsqueeze(
            0
        )  # shape: (batch, 1) * (1, dim/2) -> (batch, dim/2)
        emb = torch.cat((emb.sin(), emb.cos()), dim=-1)  # shape: (batch, dim)
        return emb


class RMSNorm(nn.Module):
    """do normalization over channel dimension \
        and multiply with a learnable parameter g"""

    def __init__(self, dim: int):
        super().__init__()
        self.g = nn.Parameter(torch.ones(1, 1, dim))

    def forward(self, x: Tensor) -> Tensor:
        return F.normalize(x, dim=2) * self.g * math.sqrt(x.shape[2])


class SelfAttention(nn.Module):
    r"""(batch, sequence, dim) -> (batch, sequence, dim)"""

    def __init__(self, dim: int, dim_head: int, num_head: int = 4):
        super().__init__()
        self.dim_inout = dim
        self.scale = dim_head**-0.5
        self.num_head = num_head
        self.hidden_dim = dim_head * num_head
        self.to_qkv = nn.Linear(self.dim_inout, self.hidden_dim * 3)

        self.to_out = nn.Sequential(
            nn.Linear(self.hidden_dim, self.dim_inout), RMSNorm(self.dim_inout)
        )
        self.rms_norm_q = nn.Sequential(
            Rearrange("b num_head l d -> b l (num_head d)"),
            RMSNorm(self.hidden_dim),  # in dim 2
            Rearrange(
                "b l (num_head d) -> b num_head l d", num_head=self.num_head
            ),
        )
        self.rms_norm_k = nn.Sequential(
            Rearrange("b num_head l d -> b l (num_head d)"),
            RMSNorm(self.hidden_dim),  # in dim 1
            Rearrange(
                "b l (num_head d) -> b num_head l d", num_head=self.num_head
            ),
        )

    def forward(self, x: Tensor) -> Tensor:
        qkv = self.to_qkv(x).chunk(
            3, dim=2
        )  # 3 * (batch, sequence, hidden_dim)
        q, k, v = map(
            lambda t: rearrange(
                t, "b l (num_head d) -> b num_head l d", num_head=self.num_head
            ),
            qkv,
        )  # shape: 3 * (batch, num_head, sequence, dim_head)

        # such normalization: one independent attention for each head
        q = self.rms_norm_q(q)
        k = self.rms_norm_k(k)
        q = q * self.scale  # scale

        sim = einsum(q, k, "b h l dq, b h n dq -> b h l n")
        sim = sim.softmax(dim=-1)  # over keys sequence
        out = einsum(
            sim, v, "b h l n, b h n dv -> b h dv l"
        )  # shape: (batch, num_head, sequence, dim_head)
        out = rearrange(
            out, "b h dv l -> b l (h dv)"
        )  # shape: (batch, sequence, hidden_dim)
        return self.to_out(out)  # shape: (batch, sequence, dim)


class DecoderBlock(nn.Module):
    """GPT2-style decoder block

    input: (batch, sequence, dim)
    condition: (batch, #sequence, dim)
    output: (batch, sequence, dim)
    """

    def __init__(
        self,
        dim_base: int,  # model dimension
        num_attn_head: int = 4,
        dim_head: None | int = None,  # default dim_base // num_head
        dim_feedforward: int = 2048,
        dropout: float = 0.1,  # at ff layer
        conditioning: bool = False,
    ):
        super().__init__()
        self.dim_base = dim_base
        self.num_attn_head = num_attn_head
        self.dim_head = dim_head if dim_head else dim_base // num_attn_head
        self.dim_feedforward = dim_feedforward
        self.dropout = dropout
        self.conditioning = conditioning

        self.ln_1 = nn.LayerNorm(
            dim_base, elementwise_affine=not self.conditioning, eps=1e-6
        )
        self.attn = SelfAttention(
            dim_base,
            dim_head=dim_base // self.num_attn_head,
            num_head=self.num_attn_head,
        )
        self.ln_2 = nn.LayerNorm(
            dim_base, elementwise_affine=not self.conditioning, eps=1e-6
        )

        # no cross attention compared with the GPT2
        self.mlp = nn.Sequential(
            nn.Linear(dim_base, dim_feedforward),
            nn.SiLU(),
            nn.Linear(dim_feedforward, dim_base),
            nn.Dropout(self.dropout),
        )

        # conditioning
        if self.conditioning:
            self.adaLN_modulation = nn.Sequential(
                nn.SiLU(),
                nn.Linear(dim_base, dim_base * 6),
            )

    def forward(
        self,
        x: Tensor,
        c: None | Tensor = None,
    ) -> Tensor:
        if c is not None and self.conditioning:
            cond_scale_shift_gate = self.adaLN_modulation(
                c
            )  # shape: (batch, 1, channel * 6)
            (
                scale_attn,
                shift_attn,
                gate_attn,
                scale_mlp,
                shift_mlp,
                gate_mlp,
            ) = cond_scale_shift_gate.chunk(6, dim=2)
        else:
            (
                scale_attn,
                shift_attn,
                gate_attn,
                scale_mlp,
                shift_mlp,
                gate_mlp,
            ) = [0.0, 0.0, 1.0, 0.0, 0.0, 1.0]

        # attention
        x = x + gate_attn * self.attn(
            self.ln_1(x) * (1.0 + scale_attn) + shift_attn
        )

        # mlp
        x = x + gate_mlp * self.mlp(
            self.ln_2(x) * (1.0 + scale_mlp) + shift_mlp
        )

        return x


class DecoderTransformer(nn.Module):
    """GPT2-style decoder-only transformer

    input: (batch, sequence, dim)
    condition: (batch, #sequence, dim)
    output: (batch, sequence, dim)
    """

    def __init__(
        self,
        dim_base: int,
        num_attn_head: int = 4,
        num_layer=6,
        dim_feedforward=2048,  # need to be big for transformer
        dropout=0.1,  # at ff layer
        conditioning: bool = False,  # for conditioning on t (diffusion)
    ):
        super().__init__()
        self.dim_base = dim_base
        self.num_attn_head = num_attn_head
        self.num_layer = num_layer
        self.dim_feedforward = dim_feedforward
        self.dropout = dropout
        self.conditioning = conditioning

        # layers
        self.decoders = nn.ModuleList([])
        for idx in range(self.num_layer):
            self.decoders.append(
                DecoderBlock(
                    dim_base=dim_base,
                    num_attn_head=num_attn_head,
                    dim_feedforward=dim_feedforward,
                    dropout=dropout,
                    conditioning=conditioning,
                )  # shape: (batch, sequence, dim) -> (batch, sequence, dim)
            )

    def forward(
        self,
        x: Tensor,
        c: None | Tensor = None,
    ) -> Tensor:
        for decoder in self.decoders:
            x = decoder(x, c)
        return x  # shape: (batch, sequence, dim)


class InitProjection(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, padding):
        super().__init__()
        self.conv = nn.Conv1d(
            in_channels, out_channels, kernel_size=kernel_size, padding=padding
        )

    def forward(self, x: Tensor) -> Tensor:
        x = x.permute(0, 2, 1)
        x = self.conv(x)
        x = x.permute(0, 2, 1)
        return x


class DenoisingTransformer(nn.Module):
    """wraps DecoderTransformer with additional layers for denoising
    input:
        - x: (batch, sequence, dim) # TODO maybe adapt?
        - time: (batch, )

    return:
        - x: (batch, sequence, dim)

    """

    def __init__(
        self,
        dim_base: int,
        dim_in: int,  # input dimension (batch, ..., dim_in)
        num_attn_head: int = 4,
        num_decoder_layer: int = 6,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        learn_variance: bool = False,  # learnable variance
        disable_init_proj: bool = False,  # disable initial projection
    ):
        super().__init__()
        self.dim_base = dim_base
        self.dim_in = dim_in
        self.dim_out = dim_in * (2 if learn_variance else 1)
        self.num_attn_head = num_attn_head
        self.num_decoder_layer = num_decoder_layer
        self.dim_feedforward = dim_feedforward
        self.dropout = dropout
        self.learn_variance = learn_variance

        # layers
        time_pos_emb = SinusoidalPosEmb(self.dim_base)
        self.time_mlp = nn.Sequential(
            time_pos_emb,
            nn.Linear(self.dim_base, self.dim_base),
            nn.GELU(),
            nn.Linear(self.dim_base, self.dim_base * 3),
        )  # scale, shift, input_to_decoder_blocks

        _transformer_pos_emb = SinusoidalPosEmb(self.dim_base)
        self.transformer_pos_emb = nn.Sequential(
            _transformer_pos_emb,
            nn.Linear(self.dim_base, self.dim_base),
            nn.GELU(),
            nn.Linear(self.dim_base, self.dim_base),
        )

        # initial projection
        # self.init_proj = nn.Sequential(
        #     Rearrange('batch seq dim_in -> batch dim_in seq'),
        #     nn.Conv1d(self.dim_in, self.dim_base, kernel_size=5, padding=2),
        #     Rearrange('batch dim_base seq -> batch seq dim_base')
        # ) # (batch, sequence, dim_in) -> (batch, sequence, dim_base)
        if not disable_init_proj:
            self.init_proj = InitProjection(
                self.dim_in, self.dim_base, kernel_size=5, padding=2
            )
        else:
            self.init_proj = nn.Linear(self.dim_in, self.dim_base)

        # decoder
        self.transformer = DecoderTransformer(
            dim_base=dim_base,
            num_attn_head=num_attn_head,
            num_layer=num_decoder_layer,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            conditioning=True,  # on timestep t
        )

        # conditioning
        self.final_ln = nn.LayerNorm(
            self.dim_base, elementwise_affine=False, eps=1e-6
        )
        self.final_adaLN_modulation = nn.Sequential(
            nn.SiLU(),
            nn.Linear(self.dim_base, self.dim_base * 2),
        )

        # final linear
        self.final_linear = nn.Linear(self.dim_base * 2, self.dim_out)

        # intialize weights
        self.initialize_weights()

    def initialize_weights(self):
        def _basic_init(module):
            if getattr(module, "weight", None) is not None and not isinstance(
                module, nn.LayerNorm
            ):
                nn.init.xavier_uniform_(module.weight)
            if getattr(module, "bias", None) is not None:
                nn.init.constant_(module.bias, 0)

        self.apply(_basic_init)

        # better not zero out the weights of final output layer
        # the final linear: (dim_base * 2) -> (dim_out)
        nn.init.constant_(self.final_linear.bias, 0)

        for decoder in self.transformer.decoders:
            nn.init.constant_(decoder.adaLN_modulation[-1].weight, 0)
            nn.init.constant_(decoder.adaLN_modulation[-1].bias, 0)

        nn.init.constant_(self.final_adaLN_modulation[-1].weight, 0)
        nn.init.constant_(self.final_adaLN_modulation[-1].bias, 0)

    def forward(
        self,
        x: Tensor,
        time: Tensor,
    ) -> Tensor:
        # encoder time
        _encoded_t = self.time_mlp(time)  # shape: (batch, dim_base * 3)
        _encoded_t = rearrange(_encoded_t, "batch dim -> batch 1 dim")
        scale, shift, encoded_t = _encoded_t.chunk(
            3, dim=2
        )  # shape: 3 * (batch, 1, dim_base)

        # init projection
        x = self.init_proj(x)  # shape: (batch, sequence, dim_base)
        x_copy = x.clone()
        x = x * (1 + scale) + shift  # shape: (batch, sequence, dim_base)
        x = F.silu(x)  # shape: (batch, sequence, dim_base)

        # transformer
        pos_seq = torch.arange(
            x.shape[1],
            device=x.device,
            dtype=x.dtype,
        )
        pos_emb_seq = self.transformer_pos_emb(
            pos_seq
        )  # shape: (sequence, dim_base)
        pos_emb_seq = rearrange(
            pos_emb_seq, "seq dim -> 1 seq dim"
        )  # shape: (1, sequence, dim_base)
        x = x + pos_emb_seq

        x = self.transformer(
            x, encoded_t
        )  # shape: (batch, sequence, dim_base)

        # final adaln modulation
        scale_final, shift_final = self.final_adaLN_modulation(
            encoded_t
        ).chunk(
            2, dim=2
        )  # shape: 2 * (batch, 1, dim_base)
        x = self.final_ln(x) * (1 + scale_final) + shift_final

        # final linear
        x = torch.concat(
            (x, x_copy), dim=2
        )  # shape: (batch, sequence, dim_base * 2)
        x = self.final_linear(x)

        return x  # shape: (batch, sequence, dim_out)



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0
# Disclaimer
#   part of this script uses
# DPM-Solver:https://github.com/LuChengTHU/dpm-solver
"""
Advised to only directly use `DPMSolverSampler` in this module.
"""
import math
from typing import Tuple

import torch
from torch import Tensor
from tqdm.auto import tqdm

from opensynth.models.energydiff._diffusion_base import (
    BetaScheduleType,
    DiffusionBase,
    ModelMeanType,
    ModelVarianceType,
)


class NoiseScheduleVP:
    def __init__(
        self,
        schedule=BetaScheduleType.DISCRETE,
        betas=None,
        alphas_cumprod=None,
        continuous_beta_0=0.1,
        continuous_beta_1=20.0,
    ):
        if not isinstance(schedule, BetaScheduleType):
            raise ValueError("Schedule must be a BetaScheduleType.")

        self.schedule = schedule
        if schedule == BetaScheduleType.DISCRETE:
            if betas is not None:
                log_alphas = 0.5 * torch.log(1 - betas).cumsum(dim=0)
            else:
                assert alphas_cumprod is not None
                log_alphas = 0.5 * torch.log(alphas_cumprod)
            self.total_N = len(log_alphas)
            self.T = 1.0
            self.t_array = torch.linspace(0.0, 1.0, self.total_N + 1)[
                1:
            ].reshape((1, -1))
            self.log_alpha_array = log_alphas.reshape(
                (
                    1,
                    -1,
                )
            )
        else:
            self.total_N = 1000
            self.beta_0 = continuous_beta_0
            self.beta_1 = continuous_beta_1
            self.cosine_s = 0.008
            self.cosine_beta_max = 999.0
            self.cosine_t_max = (
                math.atan(
                    self.cosine_beta_max * (1.0 + self.cosine_s) / math.pi
                )
                * 2.0
                * (1.0 + self.cosine_s)
                / math.pi
                - self.cosine_s
            )
            self.cosine_log_alpha_0 = math.log(
                math.cos(self.cosine_s / (1.0 + self.cosine_s) * math.pi / 2.0)
            )
            self.schedule = schedule
            if schedule == BetaScheduleType.COSINE:
                self.T = 0.9946
            else:
                self.T = 1.0

    def marginal_log_mean_coeff(self, t: Tensor) -> Tensor:
        """
        Compute log(alpha_t) of a given continuous-time label t in [0, T].
        """
        if self.schedule == BetaScheduleType.DISCRETE:
            return interpolate_fn(
                t.reshape((-1, 1)),
                self.t_array.to(t.device),
                self.log_alpha_array.to(t.device),
            ).reshape((-1))
        elif self.schedule == BetaScheduleType.LINEAR:
            return (
                -0.25 * t**2 * (self.beta_1 - self.beta_0)
                - 0.5 * t * self.beta_0
            )
        elif self.schedule == BetaScheduleType.COSINE:

            def log_alpha_fn(s):
                return torch.log(
                    torch.cos(
                        (s + self.cosine_s)
                        / (1.0 + self.cosine_s)
                        * math.pi
                        / 2.0
                    )
                )

            log_alpha_t = log_alpha_fn(t) - self.cosine_log_alpha_0
            return log_alpha_t

    def marginal_alpha(self, t):
        """
        Compute alpha_t of a given continuous-time label t in [0, T].
        """
        return torch.exp(self.marginal_log_mean_coeff(t))

    def marginal_std(self, t):
        """
        Compute sigma_t of a given continuous-time label t in [0, T].
        """
        return torch.sqrt(
            1.0 - torch.exp(2.0 * self.marginal_log_mean_coeff(t))
        )

    def marginal_lambda(self, t):
        """
        Compute lambda_t = log(alpha_t) - log(sigma_t)
            of a given continuous-time label t in [0, T].
        """
        log_mean_coeff = self.marginal_log_mean_coeff(t)
        log_std = 0.5 * torch.log(1.0 - torch.exp(2.0 * log_mean_coeff))
        return log_mean_coeff - log_std

    def inverse_lambda(self, lamb):
        """
        Compute the continuous-time label t
            in [0, T] of a given half-logSNR lambda_t.
        """
        if self.schedule == BetaScheduleType.LINEAR:
            tmp = (
                2.0
                * (self.beta_1 - self.beta_0)
                * torch.logaddexp(-2.0 * lamb, torch.zeros((1,)).to(lamb))
            )
            Delta = self.beta_0**2 + tmp
            return (
                tmp
                / (torch.sqrt(Delta) + self.beta_0)
                / (self.beta_1 - self.beta_0)
            )
        elif self.schedule == BetaScheduleType.DISCRETE:
            log_alpha = -0.5 * torch.logaddexp(
                torch.zeros((1,)).to(lamb.device), -2.0 * lamb
            )
            t = interpolate_fn(
                log_alpha.reshape((-1, 1)),
                torch.flip(self.log_alpha_array.to(lamb.device), [1]),
                torch.flip(self.t_array.to(lamb.device), [1]),
            )
            return t.reshape((-1,))
        else:
            log_alpha = -0.5 * torch.logaddexp(
                -2.0 * lamb, torch.zeros((1,)).to(lamb)
            )
            t_fn = (
                lambda log_alpha_t: torch.arccos(
                    torch.exp(log_alpha_t + self.cosine_log_alpha_0)
                )
                * 2.0
                * (1.0 + self.cosine_s)
                / math.pi
                - self.cosine_s
            )
            t = t_fn(log_alpha)
            return t


def model_wrapper(
    model,
    noise_schedule,
    model_type="noise",
    model_kwargs={},
    guidance_type="uncond",
    condition=None,
    unconditional_condition=None,
    cfg_scale=1.0,
    classifier_fn=None,
    classifier_kwargs={},
):
    def get_model_input_time(t_continuous):
        """
        Convert the continuous-time `t_continuous`
            (in [epsilon, T]) to the model input time.
        For discrete-time DPMs,
            we convert `t_continuous` in [1 / N, 1] to `t_input`
                in [0, 1000 * (N - 1) / N].
        For continuous-time DPMs, we just use `t_continuous`.
        """
        if noise_schedule.schedule == BetaScheduleType.DISCRETE:
            return (t_continuous - 1.0 / noise_schedule.total_N) * 1000.0
        else:
            return t_continuous

    def noise_pred_fn(x, t_continuous, cond=None):
        if t_continuous.reshape((-1,)).shape[0] == 1:
            t_continuous = t_continuous.expand((x.shape[0]))
        t_input = get_model_input_time(t_continuous)
        if cond is None:
            output = model(x, t_input, **model_kwargs)
        else:
            output = model(x, t_input, cond, **model_kwargs)
        if model_type == "noise":
            return output
        elif model_type == "x_start":
            alpha_t, sigma_t = noise_schedule.marginal_alpha(
                t_continuous
            ), noise_schedule.marginal_std(t_continuous)
            dims = x.dim()
            return (x - expand_dims(alpha_t, dims) * output) / expand_dims(
                sigma_t, dims
            )
        elif model_type == "v":
            alpha_t, sigma_t = noise_schedule.marginal_alpha(
                t_continuous
            ), noise_schedule.marginal_std(t_continuous)
            dims = x.dim()
            return (
                expand_dims(alpha_t, dims) * output
                + expand_dims(sigma_t, dims) * x
            )
        elif model_type == "score":
            sigma_t = noise_schedule.marginal_std(t_continuous)
            dims = x.dim()
            return -expand_dims(sigma_t, dims) * output

    def cond_grad_fn(x, t_input):
        """
        Compute the gradient of the classifier,
            i.e. nabla_{x} log p_t(cond | x_t).
        """
        with torch.enable_grad():
            x_in = x.detach().requires_grad_(True)
            log_prob = classifier_fn(
                x_in, t_input, condition, **classifier_kwargs
            )
            return torch.autograd.grad(log_prob.sum(), x_in)[0]

    def model_fn(x, t_continuous):
        """
        The noise predicition model function that is used for DPM-Solver.
        """
        if t_continuous.reshape((-1,)).shape[0] == 1:
            t_continuous = t_continuous.expand((x.shape[0]))
        if guidance_type == "uncond":
            return noise_pred_fn(x, t_continuous)
        elif guidance_type == "classifier":
            assert classifier_fn is not None
            t_input = get_model_input_time(t_continuous)
            cond_grad = cond_grad_fn(x, t_input)
            sigma_t = noise_schedule.marginal_std(t_continuous)
            noise = noise_pred_fn(x, t_continuous)
            return (
                noise
                - cfg_scale
                * expand_dims(sigma_t, dims=cond_grad.dim())
                * cond_grad
            )
        elif guidance_type == "classifier-free":
            if cfg_scale == 1.0 or unconditional_condition is None:
                return noise_pred_fn(x, t_continuous, cond=condition)
            else:
                x_in = torch.cat([x] * 2)
                t_in = torch.cat([t_continuous] * 2)
                c_in = torch.cat([unconditional_condition, condition])
                noise_uncond, noise = noise_pred_fn(
                    x_in, t_in, cond=c_in
                ).chunk(2)
                return noise_uncond + cfg_scale * (noise - noise_uncond)

    assert model_type in ["noise", "x_start", "v"]
    assert guidance_type in ["uncond", "classifier", "classifier-free"]
    return model_fn


class DPM_Solver:
    def __init__(
        self,
        model_fn,
        noise_schedule,
        predict_x0=False,
        thresholding=False,
        max_val=1.0,
    ):
        self.model = model_fn
        self.noise_schedule = noise_schedule
        self.predict_x0 = predict_x0
        self.thresholding = thresholding
        self.max_val = max_val

    def noise_prediction_fn(self, x, t):
        """
        Return the noise prediction model.
        """
        return self.model(x, t)

    def data_prediction_fn(self, x, t):
        """
        Return the data prediction model (with thresholding).
        """
        noise = self.noise_prediction_fn(x, t)
        dims = x.dim()
        alpha_t, sigma_t = self.noise_schedule.marginal_alpha(
            t
        ), self.noise_schedule.marginal_std(t)
        x0 = (x - expand_dims(sigma_t, dims) * noise) / expand_dims(
            alpha_t, dims
        )
        if self.thresholding:
            p = 0.995  # A hyperparameter in the paper of "Imagen" [1].
            s = torch.quantile(
                torch.abs(x0).reshape((x0.shape[0], -1)), p, dim=1
            )
            s = expand_dims(
                torch.maximum(
                    s, self.max_val * torch.ones_like(s).to(s.device)
                ),
                dims,
            )
            x0 = torch.clamp(x0, -s, s) / s
        return x0

    def model_fn(self, x, t):
        """
        Convert the model to the noise prediction model
            or the data prediction model.
        """
        if self.predict_x0:
            return self.data_prediction_fn(x, t)
        else:
            return self.noise_prediction_fn(x, t)

    def get_time_steps(self, skip_type, t_T, t_0, N, device):
        if skip_type == "logSNR":
            lambda_T = self.noise_schedule.marginal_lambda(
                torch.tensor(t_T).to(device)
            )
            lambda_0 = self.noise_schedule.marginal_lambda(
                torch.tensor(t_0).to(device)
            )
            logSNR_steps = torch.linspace(
                lambda_T.cpu().item(), lambda_0.cpu().item(), N + 1
            ).to(device)
            return self.noise_schedule.inverse_lambda(logSNR_steps)
        elif skip_type == "time_uniform":
            return torch.linspace(t_T, t_0, N + 1).to(device)
        elif skip_type == "time_quadratic":
            t_order = 2
            t = (
                torch.linspace(
                    t_T ** (1.0 / t_order), t_0 ** (1.0 / t_order), N + 1
                )
                .pow(t_order)
                .to(device)
            )
            return t
        else:
            raise ValueError(
                "Unsupported skip_type {}, need to be 'logSNR' \
                    or 'time_uniform' \
                        or 'time_quadratic'".format(
                    skip_type
                )
            )

    def get_orders_and_timesteps_for_singlestep_solver(
        self, steps, order, skip_type, t_T, t_0, device
    ):
        if order == 3:
            K = steps // 3 + 1
            if steps % 3 == 0:
                orders = [
                    3,
                ] * (
                    K - 2
                ) + [2, 1]
            elif steps % 3 == 1:
                orders = [
                    3,
                ] * (
                    K - 1
                ) + [1]
            else:
                orders = [
                    3,
                ] * (
                    K - 1
                ) + [2]
        elif order == 2:
            if steps % 2 == 0:
                K = steps // 2
                orders = [
                    2,
                ] * K
            else:
                K = steps // 2 + 1
                orders = [
                    2,
                ] * (
                    K - 1
                ) + [1]
        elif order == 1:
            K = 1
            orders = [
                1,
            ] * steps
        else:
            raise ValueError("'order' must be '1' or '2' or '3'.")
        if skip_type == "logSNR":
            # To reproduce the results in DPM-Solver paper
            timesteps_outer = self.get_time_steps(
                skip_type, t_T, t_0, K, device
            )
        else:
            timesteps_outer = self.get_time_steps(
                skip_type, t_T, t_0, steps, device
            )[
                torch.cumsum(
                    torch.tensor(
                        [
                            0,
                        ]
                        + orders
                    )
                ).to(device)
            ]
        return timesteps_outer, orders

    def denoise_to_zero_fn(self, x, s):
        """
        Denoise at the final step, which is equivalent to
            solve the ODE from lambda_s to infty
                by first-order discretization.
        """
        return self.data_prediction_fn(x, s)

    def dpm_solver_first_update(
        self, x, s, t, model_s=None, return_intermediate=False
    ):
        """
        DPM-Solver-1 (equivalent to DDIM) from time `s` to time `t`.

        Args:
            x: A pytorch tensor. The initial value at time `s`.
            s: A pytorch tensor. The starting time, with the shape
                (x.shape[0],).
            t: A pytorch tensor. The ending time,
                with the shape (x.shape[0],).
            model_s: A pytorch tensor.
                The model function evaluated at time `s`.
                If `model_s` is None, we evaluate the model
                    by `x` and `s`;
                    otherwise we directly use it.
            return_intermediate: A `bool`.
                If true, also return the model value at time `s`.
        Returns:
            x_t: A pytorch tensor. The approximated solution
                at time `t`.
        """
        ns = self.noise_schedule
        dims = x.dim()
        lambda_s, lambda_t = ns.marginal_lambda(s), ns.marginal_lambda(t)
        h = lambda_t - lambda_s
        log_alpha_s, log_alpha_t = ns.marginal_log_mean_coeff(
            s
        ), ns.marginal_log_mean_coeff(t)
        sigma_s, sigma_t = ns.marginal_std(s), ns.marginal_std(t)
        alpha_t = torch.exp(log_alpha_t)

        if self.predict_x0:
            phi_1 = torch.expm1(-h)
            if model_s is None:
                model_s = self.model_fn(x, s)
            x_t = (
                expand_dims(sigma_t / sigma_s, dims) * x
                - expand_dims(alpha_t * phi_1, dims) * model_s
            )
            if return_intermediate:
                return x_t, {"model_s": model_s}
            else:
                return x_t
        else:
            phi_1 = torch.expm1(h)
            if model_s is None:
                model_s = self.model_fn(x, s)
            x_t = (
                expand_dims(torch.exp(log_alpha_t - log_alpha_s), dims) * x
                - expand_dims(sigma_t * phi_1, dims) * model_s
            )
            if return_intermediate:
                return x_t, {"model_s": model_s}
            else:
                return x_t

    def singlestep_dpm_solver_second_update(
        self,
        x,
        s,
        t,
        r1=0.5,
        model_s=None,
        return_intermediate=False,
        solver_type="dpm_solver",
    ):
        if solver_type not in ["dpm_solver", "taylor"]:
            raise ValueError(
                "'solver_type' must be either 'dpm_solver' \
                        or 'taylor', got {}".format(
                    solver_type
                )
            )
        if r1 is None:
            r1 = 0.5
        ns = self.noise_schedule
        dims = x.dim()
        lambda_s, lambda_t = ns.marginal_lambda(s), ns.marginal_lambda(t)
        h = lambda_t - lambda_s
        lambda_s1 = lambda_s + r1 * h
        s1 = ns.inverse_lambda(lambda_s1)
        log_alpha_s, log_alpha_s1, log_alpha_t = (
            ns.marginal_log_mean_coeff(s),
            ns.marginal_log_mean_coeff(s1),
            ns.marginal_log_mean_coeff(t),
        )
        sigma_s, sigma_s1, sigma_t = (
            ns.marginal_std(s),
            ns.marginal_std(s1),
            ns.marginal_std(t),
        )
        alpha_s1, alpha_t = torch.exp(log_alpha_s1), torch.exp(log_alpha_t)

        if self.predict_x0:
            phi_11 = torch.expm1(-r1 * h)
            phi_1 = torch.expm1(-h)

            if model_s is None:
                model_s = self.model_fn(x, s)
            x_s1 = (
                expand_dims(sigma_s1 / sigma_s, dims) * x
                - expand_dims(alpha_s1 * phi_11, dims) * model_s
            )
            model_s1 = self.model_fn(x_s1, s1)
            if solver_type == "dpm_solver":
                x_t = (
                    expand_dims(sigma_t / sigma_s, dims) * x
                    - expand_dims(alpha_t * phi_1, dims) * model_s
                    - (0.5 / r1)
                    * expand_dims(alpha_t * phi_1, dims)
                    * (model_s1 - model_s)
                )
            elif solver_type == "taylor":
                x_t = (
                    expand_dims(sigma_t / sigma_s, dims) * x
                    - expand_dims(alpha_t * phi_1, dims) * model_s
                    + (1.0 / r1)
                    * expand_dims(
                        alpha_t * ((torch.exp(-h) - 1.0) / h + 1.0), dims
                    )
                    * (model_s1 - model_s)
                )
        else:
            phi_11 = torch.expm1(r1 * h)
            phi_1 = torch.expm1(h)

            if model_s is None:
                model_s = self.model_fn(x, s)
            x_s1 = (
                expand_dims(torch.exp(log_alpha_s1 - log_alpha_s), dims) * x
                - expand_dims(sigma_s1 * phi_11, dims) * model_s
            )
            model_s1 = self.model_fn(x_s1, s1)
            if solver_type == "dpm_solver":
                x_t = (
                    expand_dims(torch.exp(log_alpha_t - log_alpha_s), dims) * x
                    - expand_dims(sigma_t * phi_1, dims) * model_s
                    - (0.5 / r1)
                    * expand_dims(sigma_t * phi_1, dims)
                    * (model_s1 - model_s)
                )
            elif solver_type == "taylor":
                x_t = (
                    expand_dims(torch.exp(log_alpha_t - log_alpha_s), dims) * x
                    - expand_dims(sigma_t * phi_1, dims) * model_s
                    - (1.0 / r1)
                    * expand_dims(
                        sigma_t * ((torch.exp(h) - 1.0) / h - 1.0), dims
                    )
                    * (model_s1 - model_s)
                )
        if return_intermediate:
            return x_t, {"model_s": model_s, "model_s1": model_s1}
        else:
            return x_t

    def singlestep_dpm_solver_third_update(
        self,
        x,
        s,
        t,
        r1=1.0 / 3.0,
        r2=2.0 / 3.0,
        model_s=None,
        model_s1=None,
        return_intermediate=False,
        solver_type="dpm_solver",
    ):
        if solver_type not in ["dpm_solver", "taylor"]:
            raise ValueError(
                "'solver_type' must be either 'dpm_solver' or 'taylor', \
                    got {}".format(
                    solver_type
                )
            )
        if r1 is None:
            r1 = 1.0 / 3.0
        if r2 is None:
            r2 = 2.0 / 3.0
        ns = self.noise_schedule
        dims = x.dim()
        lambda_s, lambda_t = ns.marginal_lambda(s), ns.marginal_lambda(t)
        h = lambda_t - lambda_s
        lambda_s1 = lambda_s + r1 * h
        lambda_s2 = lambda_s + r2 * h
        s1 = ns.inverse_lambda(lambda_s1)
        s2 = ns.inverse_lambda(lambda_s2)
        log_alpha_s, log_alpha_s1, log_alpha_s2, log_alpha_t = (
            ns.marginal_log_mean_coeff(s),
            ns.marginal_log_mean_coeff(s1),
            ns.marginal_log_mean_coeff(s2),
            ns.marginal_log_mean_coeff(t),
        )
        sigma_s, sigma_s1, sigma_s2, sigma_t = (
            ns.marginal_std(s),
            ns.marginal_std(s1),
            ns.marginal_std(s2),
            ns.marginal_std(t),
        )
        alpha_s1, alpha_s2, alpha_t = (
            torch.exp(log_alpha_s1),
            torch.exp(log_alpha_s2),
            torch.exp(log_alpha_t),
        )

        if self.predict_x0:
            phi_11 = torch.expm1(-r1 * h)
            phi_12 = torch.expm1(-r2 * h)
            phi_1 = torch.expm1(-h)
            phi_22 = torch.expm1(-r2 * h) / (r2 * h) + 1.0
            phi_2 = phi_1 / h + 1.0
            phi_3 = phi_2 / h - 0.5

            if model_s is None:
                model_s = self.model_fn(x, s)
            if model_s1 is None:
                x_s1 = (
                    expand_dims(sigma_s1 / sigma_s, dims) * x
                    - expand_dims(alpha_s1 * phi_11, dims) * model_s
                )
                model_s1 = self.model_fn(x_s1, s1)
            x_s2 = (
                expand_dims(sigma_s2 / sigma_s, dims) * x
                - expand_dims(alpha_s2 * phi_12, dims) * model_s
                + r2
                / r1
                * expand_dims(alpha_s2 * phi_22, dims)
                * (model_s1 - model_s)
            )
            model_s2 = self.model_fn(x_s2, s2)
            if solver_type == "dpm_solver":
                x_t = (
                    expand_dims(sigma_t / sigma_s, dims) * x
                    - expand_dims(alpha_t * phi_1, dims) * model_s
                    + (1.0 / r2)
                    * expand_dims(alpha_t * phi_2, dims)
                    * (model_s2 - model_s)
                )
            elif solver_type == "taylor":
                D1_0 = (1.0 / r1) * (model_s1 - model_s)
                D1_1 = (1.0 / r2) * (model_s2 - model_s)
                D1 = (r2 * D1_0 - r1 * D1_1) / (r2 - r1)
                D2 = 2.0 * (D1_1 - D1_0) / (r2 - r1)
                x_t = (
                    expand_dims(sigma_t / sigma_s, dims) * x
                    - expand_dims(alpha_t * phi_1, dims) * model_s
                    + expand_dims(alpha_t * phi_2, dims) * D1
                    - expand_dims(alpha_t * phi_3, dims) * D2
                )
        else:
            phi_11 = torch.expm1(r1 * h)
            phi_12 = torch.expm1(r2 * h)
            phi_1 = torch.expm1(h)
            phi_22 = torch.expm1(r2 * h) / (r2 * h) - 1.0
            phi_2 = phi_1 / h - 1.0
            phi_3 = phi_2 / h - 0.5

            if model_s is None:
                model_s = self.model_fn(x, s)
            if model_s1 is None:
                x_s1 = (
                    expand_dims(torch.exp(log_alpha_s1 - log_alpha_s), dims)
                    * x
                    - expand_dims(sigma_s1 * phi_11, dims) * model_s
                )
                model_s1 = self.model_fn(x_s1, s1)
            x_s2 = (
                expand_dims(torch.exp(log_alpha_s2 - log_alpha_s), dims) * x
                - expand_dims(sigma_s2 * phi_12, dims) * model_s
                - r2
                / r1
                * expand_dims(sigma_s2 * phi_22, dims)
                * (model_s1 - model_s)
            )
            model_s2 = self.model_fn(x_s2, s2)
            if solver_type == "dpm_solver":
                x_t = (
                    expand_dims(torch.exp(log_alpha_t - log_alpha_s), dims) * x
                    - expand_dims(sigma_t * phi_1, dims) * model_s
                    - (1.0 / r2)
                    * expand_dims(sigma_t * phi_2, dims)
                    * (model_s2 - model_s)
                )
            elif solver_type == "taylor":
                D1_0 = (1.0 / r1) * (model_s1 - model_s)
                D1_1 = (1.0 / r2) * (model_s2 - model_s)
                D1 = (r2 * D1_0 - r1 * D1_1) / (r2 - r1)
                D2 = 2.0 * (D1_1 - D1_0) / (r2 - r1)
                x_t = (
                    expand_dims(torch.exp(log_alpha_t - log_alpha_s), dims) * x
                    - expand_dims(sigma_t * phi_1, dims) * model_s
                    - expand_dims(sigma_t * phi_2, dims) * D1
                    - expand_dims(sigma_t * phi_3, dims) * D2
                )

        if return_intermediate:
            return x_t, {
                "model_s": model_s,
                "model_s1": model_s1,
                "model_s2": model_s2,
            }
        else:
            return x_t

    def multistep_dpm_solver_second_update(
        self, x, model_prev_list, t_prev_list, t, solver_type="dpm_solver"
    ):
        if solver_type not in ["dpm_solver", "taylor"]:
            raise ValueError(
                "'solver_type' must be either 'dpm_solver' or \
                    'taylor', got {}".format(
                    solver_type
                )
            )
        ns = self.noise_schedule
        dims = x.dim()
        model_prev_1, model_prev_0 = model_prev_list
        t_prev_1, t_prev_0 = t_prev_list
        lambda_prev_1, lambda_prev_0, lambda_t = (
            ns.marginal_lambda(t_prev_1),
            ns.marginal_lambda(t_prev_0),
            ns.marginal_lambda(t),
        )
        log_alpha_prev_0, log_alpha_t = ns.marginal_log_mean_coeff(
            t_prev_0
        ), ns.marginal_log_mean_coeff(t)
        sigma_prev_0, sigma_t = ns.marginal_std(t_prev_0), ns.marginal_std(t)
        alpha_t = torch.exp(log_alpha_t)

        h_0 = lambda_prev_0 - lambda_prev_1
        h = lambda_t - lambda_prev_0
        r0 = h_0 / h
        D1_0 = expand_dims(1.0 / r0, dims) * (model_prev_0 - model_prev_1)
        if self.predict_x0:
            if solver_type == "dpm_solver":
                x_t = (
                    expand_dims(sigma_t / sigma_prev_0, dims) * x
                    - expand_dims(alpha_t * (torch.exp(-h) - 1.0), dims)
                    * model_prev_0
                    - 0.5
                    * expand_dims(alpha_t * (torch.exp(-h) - 1.0), dims)
                    * D1_0
                )
            elif solver_type == "taylor":
                x_t = (
                    expand_dims(sigma_t / sigma_prev_0, dims) * x
                    - expand_dims(alpha_t * (torch.exp(-h) - 1.0), dims)
                    * model_prev_0
                    + expand_dims(
                        alpha_t * ((torch.exp(-h) - 1.0) / h + 1.0), dims
                    )
                    * D1_0
                )
        else:
            if solver_type == "dpm_solver":
                x_t = (
                    expand_dims(
                        torch.exp(log_alpha_t - log_alpha_prev_0), dims
                    )
                    * x
                    - expand_dims(sigma_t * (torch.exp(h) - 1.0), dims)
                    * model_prev_0
                    - 0.5
                    * expand_dims(sigma_t * (torch.exp(h) - 1.0), dims)
                    * D1_0
                )
            elif solver_type == "taylor":
                x_t = (
                    expand_dims(
                        torch.exp(log_alpha_t - log_alpha_prev_0), dims
                    )
                    * x
                    - expand_dims(sigma_t * (torch.exp(h) - 1.0), dims)
                    * model_prev_0
                    - expand_dims(
                        sigma_t * ((torch.exp(h) - 1.0) / h - 1.0), dims
                    )
                    * D1_0
                )
        return x_t

    def multistep_dpm_solver_third_update(
        self, x, model_prev_list, t_prev_list, t, solver_type="dpm_solver"
    ):
        ns = self.noise_schedule
        dims = x.dim()
        model_prev_2, model_prev_1, model_prev_0 = model_prev_list
        t_prev_2, t_prev_1, t_prev_0 = t_prev_list
        lambda_prev_2, lambda_prev_1, lambda_prev_0, lambda_t = (
            ns.marginal_lambda(t_prev_2),
            ns.marginal_lambda(t_prev_1),
            ns.marginal_lambda(t_prev_0),
            ns.marginal_lambda(t),
        )
        log_alpha_prev_0, log_alpha_t = ns.marginal_log_mean_coeff(
            t_prev_0
        ), ns.marginal_log_mean_coeff(t)
        sigma_prev_0, sigma_t = ns.marginal_std(t_prev_0), ns.marginal_std(t)
        alpha_t = torch.exp(log_alpha_t)

        h_1 = lambda_prev_1 - lambda_prev_2
        h_0 = lambda_prev_0 - lambda_prev_1
        h = lambda_t - lambda_prev_0
        r0, r1 = h_0 / h, h_1 / h
        D1_0 = expand_dims(1.0 / r0, dims) * (model_prev_0 - model_prev_1)
        D1_1 = expand_dims(1.0 / r1, dims) * (model_prev_1 - model_prev_2)
        D1 = D1_0 + expand_dims(r0 / (r0 + r1), dims) * (D1_0 - D1_1)
        D2 = expand_dims(1.0 / (r0 + r1), dims) * (D1_0 - D1_1)
        if self.predict_x0:
            x_t = (
                expand_dims(sigma_t / sigma_prev_0, dims) * x
                - expand_dims(alpha_t * (torch.exp(-h) - 1.0), dims)
                * model_prev_0
                + expand_dims(
                    alpha_t * ((torch.exp(-h) - 1.0) / h + 1.0), dims
                )
                * D1
                - expand_dims(
                    alpha_t * ((torch.exp(-h) - 1.0 + h) / h**2 - 0.5), dims
                )
                * D2
            )
        else:
            x_t = (
                expand_dims(torch.exp(log_alpha_t - log_alpha_prev_0), dims)
                * x
                - expand_dims(sigma_t * (torch.exp(h) - 1.0), dims)
                * model_prev_0
                - expand_dims(sigma_t * ((torch.exp(h) - 1.0) / h - 1.0), dims)
                * D1
                - expand_dims(
                    sigma_t * ((torch.exp(h) - 1.0 - h) / h**2 - 0.5), dims
                )
                * D2
            )
        return x_t

    def singlestep_dpm_solver_update(
        self,
        x,
        s,
        t,
        order,
        return_intermediate=False,
        solver_type="dpm_solver",
        r1=None,
        r2=None,
    ):
        if order == 1:
            return self.dpm_solver_first_update(
                x, s, t, return_intermediate=return_intermediate
            )
        elif order == 2:
            return self.singlestep_dpm_solver_second_update(
                x,
                s,
                t,
                return_intermediate=return_intermediate,
                solver_type=solver_type,
                r1=r1,
            )
        elif order == 3:
            return self.singlestep_dpm_solver_third_update(
                x,
                s,
                t,
                return_intermediate=return_intermediate,
                solver_type=solver_type,
                r1=r1,
                r2=r2,
            )
        else:
            raise ValueError(
                "Solver order must be 1 or 2 or 3, got {}".format(order)
            )

    def multistep_dpm_solver_update(
        self,
        x,
        model_prev_list,
        t_prev_list,
        t,
        order,
        solver_type="dpm_solver",
    ):
        if order == 1:
            return self.dpm_solver_first_update(
                x, t_prev_list[-1], t, model_s=model_prev_list[-1]
            )
        elif order == 2:
            return self.multistep_dpm_solver_second_update(
                x, model_prev_list, t_prev_list, t, solver_type=solver_type
            )
        elif order == 3:
            return self.multistep_dpm_solver_third_update(
                x, model_prev_list, t_prev_list, t, solver_type=solver_type
            )
        else:
            raise ValueError(
                "Solver order must be 1 or 2 or 3, got {}".format(order)
            )

    def dpm_solver_adaptive(
        self,
        x,
        order,
        t_T,
        t_0,
        h_init=0.05,
        atol=0.0078,
        rtol=0.05,
        theta=0.9,
        t_err=1e-5,
        solver_type="dpm_solver",
    ):
        ns = self.noise_schedule
        s = t_T * torch.ones((x.shape[0],)).to(x)
        lambda_s = ns.marginal_lambda(s)
        lambda_0 = ns.marginal_lambda(t_0 * torch.ones_like(s).to(x))
        h = h_init * torch.ones_like(s).to(x)
        x_prev = x
        nfe = 0
        if order == 2:
            r1 = 0.5

            def lower_update(x, s, t):
                return self.dpm_solver_first_update(
                    x, s, t, return_intermediate=True
                )

            def higher_update(x, s, t, **kwargs):
                return self.singlestep_dpm_solver_second_update(
                    x, s, t, r1=r1, solver_type=solver_type, **kwargs
                )

        elif order == 3:
            r1, r2 = 1.0 / 3.0, 2.0 / 3.0
            lower_update = (
                lambda x, s, t: self.singlestep_dpm_solver_second_update(
                    x,
                    s,
                    t,
                    r1=r1,
                    return_intermediate=True,
                    solver_type=solver_type,
                )
            )

            def higher_update(x, s, t, **kwargs):
                return self.singlestep_dpm_solver_third_update(
                    x, s, t, r1=r1, r2=r2, solver_type=solver_type, **kwargs
                )  # noqa: E731

        else:
            raise ValueError(
                "For adaptive step size solver, \
                        order must be 2 or 3, got {}".format(
                    order
                )
            )
        while torch.abs((s - t_0)).mean() > t_err:
            t = ns.inverse_lambda(lambda_s + h)
            x_lower, lower_noise_kwargs = lower_update(x, s, t)
            x_higher = higher_update(x, s, t, **lower_noise_kwargs)
            delta = torch.max(
                torch.ones_like(x).to(x) * atol,
                rtol * torch.max(torch.abs(x_lower), torch.abs(x_prev)),
            )

            def norm_fn(v):
                return torch.sqrt(
                    torch.square(v.reshape((v.shape[0], -1))).mean(
                        dim=-1, keepdim=True
                    )
                )  # noqa: E731

            E = norm_fn((x_higher - x_lower) / delta).max()
            if torch.all(E <= 1.0):
                x = x_higher
                s = t
                x_prev = x_lower
                lambda_s = ns.marginal_lambda(s)
            h = torch.min(
                theta * h * torch.float_power(E, -1.0 / order).float(),
                lambda_0 - lambda_s,
            )
            nfe += order
        print("adaptive solver nfe", nfe)
        return x

    def sample(
        self,
        x,
        steps=20,
        t_start=None,
        t_end=None,
        order=3,
        skip_type="time_uniform",
        method="singlestep",
        lower_order_final=True,
        denoise_to_zero=False,
        solver_type="dpm_solver",
        atol=0.0078,
        rtol=0.05,
    ):
        t_0 = 1.0 / self.noise_schedule.total_N if t_end is None else t_end
        t_T = self.noise_schedule.T if t_start is None else t_start
        device = x.device
        if method == "adaptive":
            print("adaptive dpm solver sampling...")
            with torch.no_grad():
                x = self.dpm_solver_adaptive(
                    x,
                    order=order,
                    t_T=t_T,
                    t_0=t_0,
                    atol=atol,
                    rtol=rtol,
                    solver_type=solver_type,
                )
        elif method == "multistep":
            assert steps >= order
            timesteps = self.get_time_steps(
                skip_type=skip_type, t_T=t_T, t_0=t_0, N=steps, device=device
            )
            assert timesteps.shape[0] - 1 == steps
            with torch.no_grad():
                vec_t = timesteps[0].expand((x.shape[0]))
                model_prev_list = [self.model_fn(x, vec_t)]
                t_prev_list = [vec_t]
                # Init the first `order` values
                # by lower order multistep DPM-Solver.
                for init_order in range(1, order):
                    vec_t = timesteps[init_order].expand(x.shape[0])
                    x = self.multistep_dpm_solver_update(
                        x,
                        model_prev_list,
                        t_prev_list,
                        vec_t,
                        init_order,
                        solver_type=solver_type,
                    )
                    model_prev_list.append(self.model_fn(x, vec_t))
                    t_prev_list.append(vec_t)
                # Compute the remaining values
                # by `order`-th order multistep DPM-Solver.
                for step in tqdm(
                    range(order, steps + 1), desc="multistep_dpm_solver_update"
                ):
                    vec_t = timesteps[step].expand(x.shape[0])
                    if lower_order_final and steps < 15:
                        step_order = min(order, steps + 1 - step)
                    else:
                        step_order = order
                    x = self.multistep_dpm_solver_update(
                        x,
                        model_prev_list,
                        t_prev_list,
                        vec_t,
                        step_order,
                        solver_type=solver_type,
                    )
                    for i in range(order - 1):
                        t_prev_list[i] = t_prev_list[i + 1]
                        model_prev_list[i] = model_prev_list[i + 1]
                    t_prev_list[-1] = vec_t
                    # We do not need to evaluate the final model value.
                    if step < steps:
                        model_prev_list[-1] = self.model_fn(x, vec_t)
        elif method in ["singlestep", "singlestep_fixed"]:
            if method == "singlestep":
                timesteps_outer, orders = (
                    self.get_orders_and_timesteps_for_singlestep_solver(
                        steps=steps,
                        order=order,
                        skip_type=skip_type,
                        t_T=t_T,
                        t_0=t_0,
                        device=device,
                    )
                )
            elif method == "singlestep_fixed":
                K = steps // order
                orders = [
                    order,
                ] * K
                timesteps_outer = self.get_time_steps(
                    skip_type=skip_type, t_T=t_T, t_0=t_0, N=K, device=device
                )
            for i, order in tqdm(
                enumerate(orders), desc="singlestep_dpm_solver_update"
            ):
                t_T_inner, t_0_inner = (
                    timesteps_outer[i],
                    timesteps_outer[i + 1],
                )
                timesteps_inner = self.get_time_steps(
                    skip_type=skip_type,
                    t_T=t_T_inner.item(),
                    t_0=t_0_inner.item(),
                    N=order,
                    device=device,
                )
                lambda_inner = self.noise_schedule.marginal_lambda(
                    timesteps_inner
                )
                vec_s, vec_t = t_T_inner.tile(x.shape[0]), t_0_inner.tile(
                    x.shape[0]
                )
                h = lambda_inner[-1] - lambda_inner[0]
                r1 = (
                    None
                    if order <= 1
                    else (lambda_inner[1] - lambda_inner[0]) / h
                )
                r2 = (
                    None
                    if order <= 2
                    else (lambda_inner[2] - lambda_inner[0]) / h
                )
                x = self.singlestep_dpm_solver_update(
                    x,
                    vec_s,
                    vec_t,
                    order,
                    solver_type=solver_type,
                    r1=r1,
                    r2=r2,
                )
        if denoise_to_zero:
            x = self.denoise_to_zero_fn(
                x, torch.ones((x.shape[0],)).to(device) * t_0
            )
        return x


#############################################################
# other utility functions
#############################################################


def interpolate_fn(x, xp, yp):
    N, K = x.shape[0], xp.shape[1]
    all_x = torch.cat(
        [x.unsqueeze(2), xp.unsqueeze(0).repeat((N, 1, 1))], dim=2
    )
    sorted_all_x, x_indices = torch.sort(all_x, dim=2)
    x_idx = torch.argmin(x_indices, dim=2)
    cand_start_idx = x_idx - 1
    start_idx = torch.where(
        torch.eq(x_idx, 0),
        torch.tensor(1, device=x.device),
        torch.where(
            torch.eq(x_idx, K),
            torch.tensor(K - 2, device=x.device),
            cand_start_idx,
        ),
    )
    end_idx = torch.where(
        torch.eq(start_idx, cand_start_idx), start_idx + 2, start_idx + 1
    )
    start_x = torch.gather(
        sorted_all_x, dim=2, index=start_idx.unsqueeze(2)
    ).squeeze(2)
    end_x = torch.gather(
        sorted_all_x, dim=2, index=end_idx.unsqueeze(2)
    ).squeeze(2)
    start_idx2 = torch.where(
        torch.eq(x_idx, 0),
        torch.tensor(0, device=x.device),
        torch.where(
            torch.eq(x_idx, K),
            torch.tensor(K - 2, device=x.device),
            cand_start_idx,
        ),
    )
    y_positions_expanded = yp.unsqueeze(0).expand(N, -1, -1)
    start_y = torch.gather(
        y_positions_expanded, dim=2, index=start_idx2.unsqueeze(2)
    ).squeeze(2)
    end_y = torch.gather(
        y_positions_expanded, dim=2, index=(start_idx2 + 1).unsqueeze(2)
    ).squeeze(2)
    cand = start_y + (x - start_x) * (end_y - start_y) / (end_x - start_x)
    return cand


def expand_dims(v, dims):
    return v[(...,) + (None,) * (dims - 1)]


# the sampler interfaced with Diffusion Model
class DPMSolverSampler:
    def __init__(self, model: DiffusionBase, **kwargs):
        super().__init__()
        self.model = model
        device = next(model.parameters()).device
        to_torch = (
            lambda x: x.clone().detach().to(torch.float32).to(device)
        )  # noqa: E731
        self.alphas_cumprod = to_torch(model.alpha_cumprod)

    def register_buffer(self, name, attr):
        if isinstance(attr, torch.Tensor):
            if attr.device != next(self.model.parameters()).device:
                attr = attr.to(next(self.model.parameters()).device)
        setattr(self, name, attr)

    @torch.no_grad()
    def sample(
        self,
        S: int,  # steps
        batch_size: int,  # N
        shape: Tuple[int, int],  # (C, L)
        x_T: Tensor | None = None,
    ) -> Tuple[Tensor, None]:
        """generate samples from the diffusion model using DPM-Solver.

        Args:
            S (int): number of steps to sample. has to be >= 1
            batch_size (int): batch size
            shape (tuple): shape of one sample, (num_channel, sequence_length)
            x_T (Tensor, optional): initial tensor to be denoised. \
                Defaults to None.

        Returns:
            Tuple[Tensor, None]: sampled tensor, None
        """
        # sampling
        (C, L) = shape
        size = (batch_size, C, L)

        # print(f'Data shape for DPM-Solver sampling \
        # is {size}, sampling steps {S}')

        device = next(self.model.parameters()).device
        if x_T is None:
            img = torch.randn(size, device=device)
        else:
            img = x_T

        ns = NoiseScheduleVP(
            BetaScheduleType.DISCRETE, alphas_cumprod=self.alphas_cumprod
        )

        # original model function
        def apply_model(x, t, c=None):
            out = self.model.model.forward(x, t)
            if (
                self.model.model_variance_type
                == ModelVarianceType.LEARNED_RANGE
            ):  # GaussianDiffusion1D.model_variance_type
                out = torch.split(out, out.shape[1] // 2, dim=1)[
                    0
                ]  # discard the learned variance
            return out

        # model mean type
        if self.model.model_mean_type == ModelMeanType.NOISE:
            model_type = "noise"
        elif self.model.model_mean_type == ModelMeanType.X_START:
            model_type = "x_start"
        elif self.model.model_mean_type == ModelMeanType.V:
            model_type = "v"
        else:
            raise ValueError(
                f"Unknown model mean type {self.model.model_mean_type}"
            )
        model_fn = model_wrapper(
            apply_model,
            ns,
            model_type=model_type,
            guidance_type="classifier-free",
            condition=None,
            unconditional_condition=None,
            cfg_scale=1.0,
        )

        dpm_solver = DPM_Solver(
            model_fn, ns, predict_x0=True, thresholding=False
        )
        x = dpm_solver.sample(
            img,
            steps=S,
            skip_type="time_uniform",
            method="multistep",
            order=3,
            lower_order_final=True,
        )

        return x.to(device), None



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import torch

from opensynth.data_modules.lcl_data_module import LCLDataModule
from opensynth.models.faraday import FaradayModel


def generate_synthetic_samples(
    model: FaradayModel, dm: LCLDataModule, n_samples: int
) -> torch.Tensor:
    """
    Generate synthetic samples from model.
    TODO: Add support for non-Faraday models in the future by
    creating a parent abstract class to inherit from.

    Args:
        model (FaradayModel): Model
        dm (LCLDataModule): Data module.
        n_samples (int): Number of synhetic samples to generate.

    Returns:
        torch.Tensor: Synthetic samples.
    """
    synthetic_samples = model.sample_gmm(n_samples)
    synthetic_kwh = synthetic_samples["kwh"]
    synthetic_kwh = dm.reconstruct_kwh(synthetic_kwh)
    synthetic_kwh = torch.clip(synthetic_kwh, min=0)
    return synthetic_kwh


def draw_real_data(
    dm: LCLDataModule, n_samples: int, outliers: bool = False
) -> torch.Tensor:
    """
    Draw real samples from data module

    Args:
        dm (LCLDataModule): Data module
        n_samples (int): Number of samples to draw.
        outliers (bool, optional): Return outliers instead of normal
        training data. Defaults to False.

    Returns:
        torch.Tensor: Real training samples.
    """
    if outliers:
        samples = next(iter(dm.outlier_dataloader()))
    else:
        dm.batch_size = n_samples
        samples = next(iter(dm.train_dataloader()))

    real_kwh = samples["kwh"]
    real_kwh = dm.reconstruct_kwh(real_kwh)
    real_kwh = torch.clip(real_kwh, min=0)
    return real_kwh



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

from typing import Optional, Tuple

import numpy as np
import torch
from sklearn.cluster import KMeans

from opensynth.data_modules.lcl_data_module import TrainingData
from opensynth.models.faraday.losses import _expand_samples
from opensynth.models.faraday.vae_model import FaradayVAE


def _encode_data(data: TrainingData, vae_module: FaradayVAE) -> torch.Tensor:
    """Prepare data for the GMM by encoding it with the VAE.

    Args:
        data (TrainingData): data for training GMM
        vae_module (FaradayVAE): trained VAE model

    Returns:
        torch.Tensor: encoded data for GMM
    """
    kwh = data["kwh"]
    features = data["features"]
    vae_input = vae_module.reshape_data(kwh, features)
    vae_output = vae_module.encode(vae_input)
    model_input = vae_module.reshape_data(vae_output, features)
    return model_input


def _expand_weights(data: torch.Tensor, weights: torch.Tensor) -> torch.Tensor:
    """Expand the repeat based on the training weights and shuffle the
     expanded dataset.
     Shuffling prevents consecutive samples from being the same to prevent
      convergence issues in GMM training.

    Args:
        data (torch.Tensor): data for training GMM
        weights (torch.Tensor): number of occurances of each data point in the
            dataset

    Returns:
        torch.Tensor
    """
    model_input = _expand_samples(data, weights)
    model_input = model_input[torch.randperm(model_input.size(dim=0))]
    return model_input


def prepare_data_for_training_step(
    data: TrainingData,
    vae_module: FaradayVAE,
    sample_weights_column: Optional[str] = None,
) -> torch.Tensor:

    encoded_data = _encode_data(data, vae_module)

    if sample_weights_column is not None:
        try:
            model_input = _expand_weights(
                encoded_data, data[sample_weights_column]
            )
        except KeyError:
            raise KeyError(
                f"Column {sample_weights_column} not found in input data."
            )
        return model_input
    else:
        return encoded_data


def initialise_centroids(
    X: np.array,
    n_components: int,
    init_method: str = "kmeans",
    random_state: int = 0,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Initialise centroids for GMM.
    Currently only supports K Means.

    Args:
        X (np.array): Input data
        n_components (int): Number of components
        init_method (str, optional): Initialisation method. Defaults
        to "kmeans".
        random_state (int, optional): Random state. Defaults to 0.

    Raises:
        NotImplementedError: if init_method is other than "kmeans"

    Returns:
        Tuple[torch.Tensor, torch.Tensor, torch.Tensor]: Returns the
        labels, means (centroids), and responsibilities.

        Responsibilities are the one-hot encoded tensor of each sample
        and its labels, of size N X n_components.
    """

    n_samples = len(X)
    dtype = torch.tensor(X).dtype

    if init_method == "kmeans":
        kmeans_model = KMeans(
            n_clusters=n_components, random_state=random_state
        )
        kmeans_model.fit(X)
        labels = torch.from_numpy(kmeans_model.labels_)
        means = torch.from_numpy(kmeans_model.cluster_centers_)

        responsibilities = np.zeros((n_samples, n_components))
        responsibilities[np.arange(n_samples), labels] = 1
        responsibilities = torch.from_numpy(responsibilities)
        return (
            labels.type(dtype),
            means.type(dtype),
            responsibilities.type(dtype),
        )
    else:
        raise NotImplementedError("Only kmeans is supported for now")


def is_symmetric_positive_definite(tensor):
    # Check if the tensor is symmetric
    is_symmetric = all(
        torch.allclose(tensor[i], tensor[i].T, atol=1e-5)
        for i in range(tensor.shape[0])
    )

    # Check if the tensor is positive definite by checking eigenvalues
    eigenvalues = torch.linalg.eigvals(
        tensor
    ).real  # Use only the real part of eigenvalues
    is_positive_definite = torch.all(eigenvalues > 0)

    return is_symmetric and is_positive_definite


def torch_compute_covariance(
    X: torch.Tensor,
    means: torch.Tensor,
    responsibilities: torch.Tensor,
    weights: torch.Tensor,
    reg_covar: float,
) -> torch.Tensor:
    """
    Compute the covariance matrix for each cluster.

    Args:
        X (torch.tensor): Input data
        means (torch.Tensor): Means or centroids
        responsibilities (torch.Tensor): Responsibilities
        weights (torch.Tensor): Weights
        reg_covar (float): Regularisation

    Returns:
        torch.Tensor: Covariance matrix
    """
    n_components, n_features = means.shape
    covariances = torch.empty(
        (n_components, n_features, n_features), device=X.device
    )
    # Avoid division by zero error
    means_eps = means + torch.finfo(means.dtype).eps
    for k in range(n_components):
        diff = X - means_eps[k]
        covariances[k] = (
            torch.matmul(responsibilities[:, k] * diff.T, diff) / weights[k]
        )

        # Add small regularisation
        covariances[k] = (
            covariances[k]
            + torch.eye(n_features, device=covariances.device) * reg_covar
        )

    covariances = covariances.to(device=X.device)

    return covariances


def torch_estimate_gaussian_parameters(
    X: torch.Tensor,
    responsibilities: torch.Tensor,
    reg_covar: float,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Pytorch implmentation of sklearn's
    sklearn.mixture._gaussian_mixture._estimate_gaussian_parameters

    Args:
        X (torch.Tensor): Input data
        responsibilities (torch.Tensor): Reponsibilities,
        i.e. 1-hot encoded tensor of each data and it's cluster label.
        means (torch.Tensor): Coordinate of centroids
        reg_covar (float): Covariance regularisor

    Returns:
        Tuple[torch.Tensor, torch.Tensor, torch.Tensor]: 1) cluster weights
        i.e. number of samples in each cluster and 2) cluster proba:
        % of samples in each cluster 3) covariances

    """
    # n_components, n_features = X.shape
    weights = (
        responsibilities.sum(dim=0) + torch.finfo(responsibilities.dtype).eps
    )
    # Compute new means usint updated responsibilities
    means = torch.matmul(responsibilities.T, X) / weights.reshape(-1, 1)

    covariances = torch_compute_covariance(
        X=X,
        means=means,
        responsibilities=responsibilities,
        weights=weights,
        reg_covar=reg_covar,
    )

    weights = weights / weights.sum()
    return weights, means, covariances


def torch_compute_precision_cholesky(
    covariances: torch.Tensor, reg: float = 1e-6
) -> torch.Tensor:
    """
    Pytorch implmentation of sklearn's
    sklearn.mixture._gaussian_mixture._compute_precision_cholesky
    _compute_precision_cholesky

    Args:
        covariances (torch.Tensor): Covariance matrix

    Raises:
        ValueError: Raises error if matrix is not positive determinate.

    Returns:
        torch.Tensor: Precision Cholesky
    """
    estimate_precision_error_message = (
        "Fitting the mixture model failed because some components have "
        "ill-defined empirical covariance (for instance caused by singleton "
        "or collapsed samples). Try to decrease the number of components, "
        "or increase reg_covar."
    )
    original_device = covariances.device
    if torch.device.type == "mps":
        # torch.linalg.cholesky is not supported on MPS
        # move calculation to CPU and move tensor back to device later
        covariances = covariances.to(torch.device("cpu"))

    n_components, n_features, _ = covariances.shape
    precisions_chol = torch.empty(
        (n_components, n_features, n_features), device=covariances.device
    )

    for k, covariance in enumerate(covariances):
        try:
            cov_chol = torch.linalg.cholesky(covariance, upper=False)
        except torch.linalg.LinAlgError:
            try:
                covariance_fixed = (
                    covariance
                    + torch.eye(n_features, device=covariances.device) * reg
                )
                cov_chol = torch.linalg.cholesky(covariance_fixed, upper=False)
            except torch.linalg.LinAlgError:
                print(f"Failed for {k}th covariance with reg_covar: {reg}.")
                raise ValueError(estimate_precision_error_message)
        precisions_chol[k] = torch.linalg.solve_triangular(
            cov_chol,
            torch.eye(n_features, device=covariances.device),
            upper=False,
        ).T

    precisions_chol = precisions_chol.to(device=original_device)
    return precisions_chol.type(covariances.dtype)



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import subprocess
from enum import StrEnum
from os import path
from pathlib import Path

import numpy as np
import pandas as pd
import wget


def check_redownload(file_path: Path) -> bool:
    """
    Check if file already exists.
    If it does, ask user if they want to download it again.

    Args:
        file_path (str): file path

    Returns:
        bool: True if user expresses intent
        to download it again, False otherwise
    """
    if path.exists(file_path):
        prompt = f"{file_path} already exists. Redownload? (y/n): "
        return True if input(prompt) == "y" else False
    return True


def download_data(url: str, filename: Path, unzip: bool = False):
    """
    Download data from a URL and save it to a file.
    Function will check if file already exists,
    and if does, prompt user to confirm redownload.

    Args:
        url (str): URL to download the data from
        filename (str): File path to save the downloaded data
    """
    if check_redownload(filename):
        if filename.exists():
            os.remove(filename)
        logging.info(f"Downloading data from: {url}")
        wget.download(url, str(filename))
    else:
        logging.info("Skipping download")

    if unzip:
        # Note: LCL dataset needs to be unzipped in command line with unzip
        # It's compressed using algorithms that Python's
        # zipfile module can't handle
        subprocess.run(["unzip", filename, "-d", "data/raw"])


def gaussian_noise_generator(
    mean: float, scale: float, mean_factor: float, size: tuple[int, int]
) -> np.array:
    """
    Gaussian noise generator

    Args:
        mean (float): Mean
        scale (float): Standard Deviation
        mean_factor (float): Scaling factor to scale mean by
        size (tuple[int, int]): Outlier dimension

    Returns:
        np.array: Generated outliers
    """
    noise_array = np.random.normal(
        loc=mean * mean_factor, scale=scale, size=size
    )
    return noise_array


def gamma_noise_generator(
    mean: float, scale: float, mean_factor: float, size: tuple[int, int]
) -> np.array:
    """
    Gamma noise generator

    Args:
        mean (float): Mean
        scale (float): Standard Deviation
        mean_factor (float): Scaling factor to scale mean by
        size (tuple[int, int]): Outlier dimension

    Returns:
        np.array: Generated outliers
    """
    noise_array = np.random.gamma(
        shape=mean * mean_factor, scale=scale, size=size
    )
    return noise_array


class NoiseType(StrEnum):
    GAUSSIAN = "gaussian"
    GAMMA = "gamma"


class NoiseFactory:
    def __init__(self, noise_type: NoiseType, **kwargs):
        """
        Factory class to generate noise based on the noise type

        Args:
            noise_type (NoiseType): Noise type.
        """
        self.noise_type = noise_type
        self.kwargs = kwargs

    def generate_noise(self) -> np.array:
        """
        Returns noise based on noise type

        Raises:
            ValueError: If noise type is not supported

        Returns:
            np.array:: Numpy array of random noise
        """
        if self.noise_type not in NoiseType._value2member_map_:
            raise ValueError(f"{self.noise_type} not supported!")
        elif self.noise_type == NoiseType.GAUSSIAN:
            return gaussian_noise_generator(**self.kwargs)
        elif self.noise_type == NoiseType.GAMMA:
            return gamma_noise_generator(**self.kwargs)

    def inject_noise(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Inject noise into a dataframe. First samples n_outliers
        of rows randomly from dataframe, and overwrite kWh values
        with random noise.

        Args:
            df (pd.DataFrame): Input dataframe

        Returns:
            pd.DataFrame: Dataframe with noisy outliers.
        """
        n_outliers = self.kwargs["size"][0]
        df_out = df.sample(n_outliers)
        noise_array = self.generate_noise()
        df_out["kwh"] = noise_array.tolist()
        df_out["segment"] = str(self.noise_type)
        return df_out



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0
#
# The EnergyDiff model is made available via Nan Lin and Pedro P. Vergara
# from the Delft University of Technology. Nan Lin and Pedro P. Vergara are
# funded via the ALIGN4Energy Project (with project number NWA.1389.20.251) of
# the research programme NWA ORC 2020 which is (partly) financed by the Dutch
# Research Council (NWO), The Netherland.
import enum
from typing import Tuple

import torch
from einops import rearrange


class DataShapeType(enum.Enum):
    """
    Enum class for data shape
    """

    BATCH_CHANNEL_SEQUENCE = "batch_channel_sequence"
    BATCH_SEQUENCE = "batch_sequence"
    CHANNEL_SEQUENCE = "channel_sequence"
    SEQUENCE = "sequence"


def _transform_to_target_shape(
    x: torch.Tensor,
    target_shape: DataShapeType,
    num_channel: int | None = None,
) -> torch.Tensor:
    """transform intermediate data to target shape

    Arguments:
        x -- intermediate data of shape BATCH (CHANNEL*)SEQUENCE
        target_shape -- DataShapeType
    """
    if (
        target_shape
        in [
            DataShapeType.CHANNEL_SEQUENCE,
            DataShapeType.BATCH_CHANNEL_SEQUENCE,
        ]
        and num_channel is None
    ):
        raise ValueError("num_channel should be provided")

    if target_shape == DataShapeType.BATCH_CHANNEL_SEQUENCE:
        x = rearrange(x, "b (c l) -> b c l", c=num_channel)
    elif target_shape == DataShapeType.CHANNEL_SEQUENCE:
        x = rearrange(x, "1 (c l) -> c l", c=num_channel)
    elif target_shape == DataShapeType.SEQUENCE:
        x = rearrange(x, "1 cl -> cl")

    return x


def _transform_to_process_shape(
    y: torch.Tensor, seq_length: int
) -> Tuple[torch.Tensor, DataShapeType, int | None]:
    """transform input data into process shape (BATCH SEQUENCE)

    Arguments:
        y -- torch.Tensor, of shape (batch dim1 dim2) or (batch dim1)
        seq_length -- int, equal to dim1 * dim2 (batch dim1 dim2) \
            or dim1 (batch dim1)

    Returns:
        y -- torch.Tensor of shape (batch dim)
        input_shape_type -- DataShapeType, shape of original input data y
        num_in_channel -- int|None, number of channels in the input data y
    """
    yndim = len(y.shape)
    num_in_channel = None
    if yndim == 3:
        input_shape_type = DataShapeType.BATCH_CHANNEL_SEQUENCE
        num_in_channel, _ = y.shape[1], y.shape[2]
        y = rearrange(y, "b c l -> b (c l)")
    elif yndim == 2:
        if y.shape[0] * y.shape[1] == seq_length:
            input_shape_type = DataShapeType.CHANNEL_SEQUENCE
            num_in_channel, _ = y.shape[0], y.shape[1]
            y = rearrange(y, "c l -> 1 (c l)")
        else:
            input_shape_type = DataShapeType.BATCH_SEQUENCE
            pass
    elif yndim == 1:
        input_shape_type = DataShapeType.SEQUENCE
        y = rearrange(y, "cl -> 1 cl")
    else:
        raise ValueError("Invalid input shape")

    return y, input_shape_type, num_in_channel


class MultiDimECDF:
    """
    Batched ECDF estimation.
    """

    def __init__(self, x):
        """
        Args:
            x: input data, shape (batch, channel, sequence) \
                or (batch, sequence)
                NOTE: NOT ALLOWED: (channel, sequence) or (sequence) \
                    because at you need at least a batch to calculate ECDF
        """
        # super().__init__()
        if len(x.shape) == 2:  # (batch, sequence) / (batch, channel*sequence)
            self.dim = x.shape[1]
            x = x
        else:  # (batch, channel, sequence)
            self.dims = (x.shape[1], x.shape[2])
            self.dim = x.shape[1] * x.shape[2]
            x = rearrange(x, "n d1 d2 -> n (d1 d2)")
        self.x = x
        self.num_sample = x.shape[0]
        self.x_sorted, _ = torch.sort(
            x, dim=0
        )  # expected: DataShapeType.BATCH_SEQUENCE
        # self.x_sorted = self.stable(self.x_sorted)

    @property
    def cdf(self) -> Tuple[torch.Tensor, torch.Tensor]:
        "returns x, cdf"
        x, counts = torch.unique(
            self.x_sorted.contiguous(), dim=0, sorted=True, return_counts=True
        )  # shape (n_unique, dim)
        events = torch.cumsum(counts, dim=0)  # shape (n_unique, dim)
        cdf = events.float() / self.num_sample  # shape (n_unique, dim)
        return x, cdf

    def to(self, device):
        self.x = self.x.to(device)
        self.x_sorted = self.x_sorted.to(device)
        return self

    def transform(self, x):
        """
        Args:
            x: input data, shape (b, c, l) or (b, c*l) or (c, l) or (c*l)
        (x_sorted shape) (b, c*l)
        Returns:
            y: ECDF transformed data, shape (batch_size, dim1, dim2)
                or (batch_size, dim)
        """
        x_sorted = self.x_sorted.to(x.device)
        xndim = len(x.shape)
        input_shape_type = DataShapeType.BATCH_SEQUENCE
        num_channel = None
        if xndim == 3:
            input_shape_type = DataShapeType.BATCH_CHANNEL_SEQUENCE
            num_channel, _ = x.shape[1], x.shape[2]
            x = rearrange(x, "b c l -> b (c l)")
        elif xndim == 2:
            if x.shape[0] * x.shape[1] == self.x_sorted.shape[1]:
                input_shape_type = DataShapeType.CHANNEL_SEQUENCE
                num_channel, _ = x.shape[0], x.shape[1]
                x = rearrange(x, "c l -> 1 (c l)")
            else:
                input_shape_type = DataShapeType.BATCH_SEQUENCE
                pass
        elif xndim == 1:
            input_shape_type = DataShapeType.SEQUENCE
            x = rearrange(x, "cl -> 1 cl")
        else:
            raise ValueError("Invalid input shape")

        indices = torch.searchsorted(
            x_sorted.transpose(0, 1).contiguous(),
            x.transpose(0, 1).contiguous(),
            side="right",
        ).transpose(
            0, 1
        )  # [b, c*l]
        cdf_values = indices.float() / self.num_sample

        y = torch.clamp(cdf_values, 0.0, 1.0)

        y = _transform_to_target_shape(
            y, input_shape_type, num_channel=num_channel
        )

        return y

    def inverse_transform(self, y):
        """
        Args:
            y: ECDF transformed data,
                shape (batch_size, dim1, dim2) or (batch_size, dim)
        Returns:
            x: input data, shape (batch_size, dim)
        """
        x_sorted = self.x_sorted.to(y.device)
        y, input_shape_type, num_in_channel = _transform_to_process_shape(
            y,
            seq_length=x_sorted.shape[1],
        )  # y shape: (batch_size, dim)

        y = torch.clamp(y, 0.0, 1.0)
        y_scaled = y * self.num_sample
        indices_lower = torch.floor(y_scaled).long()  # range (0, num_sample)
        xlower = torch.gather(
            x_sorted, 0, indices_lower.clamp(min=1, max=self.num_sample) - 1
        )  # range bounded by x_sorted
        x = xlower

        x = _transform_to_target_shape(x, input_shape_type, num_in_channel)

        return x

    def re_estimate(self, x):
        """
        Args:
            x: input data, shape (n_samples, dim)
        """
        self.x = self.x.to(x.device)
        if len(x.shape) == 2:
            self.x = x
            self.dim = x.shape[1]
        else:
            self.x = rearrange(x, "n d1 d2 -> n (d1 d2)")
            self.dims = (x.shape[1], x.shape[2])
            self.dim = x.shape[1] * x.shape[2]
        self.num_sample = x.shape[0]
        self.x_sorted, _ = torch.sort(x, dim=0)

    def continue_estimate(self, x):
        """
        Args:
            x: input data, shape (n_samples, dim)
        """
        self.x = self.x.to(x.device)
        if len(x.shape) == 3:
            x = rearrange(x, "n d1 d2 -> n (d1 d2)")
        self.x = torch.cat([self.x, x], dim=0)
        self.num_sample = self.x.shape[0]
        self.x_sorted, _ = torch.sort(self.x, dim=0)

    def __call__(self, x):
        """get ecdf of x"""
        x = self.transform(x)

        return x


def calibrate(
    target,  # [batch sequence]
    source,  # [batch sequence]
):
    """
    calibrate the marginals of source data to target data.
        return calibrated source data.
    """
    ecdf_source = MultiDimECDF(source)
    ecdf_target = MultiDimECDF(target)
    calibrated_source = ecdf_target.inverse_transform(
        ecdf_source.transform(source)
    )

    return calibrated_source



# Copyright Contributors to the Opensynth-energy Project.
# SPDX-License-Identifier: Apache-2.0

import ast
import logging
import sys
from dataclasses import dataclass

import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
import torch.nn as nn
import torchmetrics
from sklearn.metrics import accuracy_score, f1_score, precision_score
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader, Dataset

from opensynth.data_modules.lcl_data_module import LCLDataModule
from opensynth.datasets.low_carbon_london.preprocess_lcl import create_outliers
from opensynth.evaluation.privacy import generate_attack_data
from opensynth.models.energydiff.diffusion import PLDiffusion1D
from opensynth.models.faraday import FaradayModel

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger(__name__)


@dataclass
class MembershipInferenceAttackSamples:
    synthetic_samples: torch.Tensor
    train_samples: torch.Tensor
    holdout_samples: torch.Tensor
    outlier_seen_samples: torch.Tensor
    outlier_unseen_same_samples: torch.Tensor
    outlier_unseen_diff_samples: torch.Tensor


def _create_unseen_outliers(
    df: pd.DataFrame, mean: float, mean_factor: int = 20
) -> torch.Tensor:
    """
    Creates unseen outliers. Calls unpon the method
    opensynth.datasets.low_carbon_london.preprocess_lcl.create_outliers.
    Creates outliers based on gamma and gaussian distribution
    around the `mean` value.

    Args:
        df (pd.DataFrame): Input dataframe to create outliers from.
        mean (float): Mean value for gaussian and gamma distribution.

    Returns:
        torch.Tensor: Generated unseen outliers.
    """
    outliers = create_outliers(df, mean, mean_factor=mean_factor)
    return torch.from_numpy(np.array(outliers["kwh"].values.tolist()))


def _create_attack_samples(
    model: FaradayModel,
    dm_train: LCLDataModule,
    dm_holdout: LCLDataModule,
    n_samples: int = 20000,
    mean_factor: int = 20,
) -> MembershipInferenceAttackSamples:
    """
    Creates attack samples for membership inference attack:

    Synthetic samples: generated from the trained model.
    Train samples: real data from the training set (with
        outliers injected.)
    Holdout samples: real data from the holdout set.
    Outlier seen samples: Outliers seen during training.
    Outlier unseen same samples: Outliers not seen during training,
        but from the same distribution as seen outliers.
    Outlier unseen diff samples: Outliers not seen during training,
        but from a different distribution to the seen outliers.

    Args:
        model (FaradayModel): Model to generate synthetic samples.
        dm_train (LCLDataModule): Train data module
        dm_holdout (LCLDataModule): Holdout data module
        n_samples (int, optional): Number of samples. Defaults to 20000.
        mean_factor (int): Mean factor for outliers. Defaults to 20.

    Returns:
        MembershipInferenceAttackSamples: Attack samples
    """
    synthetic_samples = generate_attack_data.generate_synthetic_samples(
        model, dm_train, n_samples=n_samples
    )
    train_samples = generate_attack_data.draw_real_data(
        dm=dm_train, n_samples=n_samples
    )
    holdout_samples = generate_attack_data.draw_real_data(
        dm=dm_holdout, n_samples=n_samples
    )
    outlier_samples = generate_attack_data.draw_real_data(
        dm=dm_train, outliers=True, n_samples=n_samples
    )

    df_train = dm_train.dataset.df[
        dm_train.dataset.df["segment"].isna()
    ].copy()
    outlier_unseen_same_samples = _create_unseen_outliers(
        df_train, dm_train.dataset.feature_mean, mean_factor
    )
    outlier_unseen_diff_samples = _create_unseen_outliers(
        df_train, mean=0, mean_factor=mean_factor
    )

    return MembershipInferenceAttackSamples(
        synthetic_samples=synthetic_samples,
        train_samples=train_samples,
        holdout_samples=holdout_samples,
        outlier_seen_samples=outlier_samples,
        outlier_unseen_same_samples=outlier_unseen_same_samples,
        outlier_unseen_diff_samples=outlier_unseen_diff_samples,
    )


def _create_attack_samples_no_model(
    synthetic_samples: torch.Tensor,
    dm_train: LCLDataModule,
    dm_holdout: LCLDataModule,
    n_samples: int = 20000,
    mean_factor: int = 20,
) -> MembershipInferenceAttackSamples:
    """
    Creates attack samples for membership inference attack:

    Synthetic samples: generated from the trained model.
    Train samples: real data from the training set (with
        outliers injected.)
    Holdout samples: real data from the holdout set.
    Outlier seen samples: Outliers seen during training.
    Outlier unseen same samples: Outliers not seen during training,
        but from the same distribution as seen outliers.
    Outlier unseen diff samples: Outliers not seen during training,
        but from a different distribution to the seen outliers.

    Args:
        synthetic_samples (torch.Tensor): generated synthetic samples.
        dm_train (LCLDataModule): Train data module
        dm_holdout (LCLDataModule): Holdout data module
        n_samples (int, optional): Number of samples. Defaults to 20000.
        mean_factor (int): Mean factor for outliers. Defaults to 20.

    Returns:
        MembershipInferenceAttackSamples: Attack samples
    """
    train_samples = generate_attack_data.draw_real_data(
        dm=dm_train, n_samples=n_samples
    )
    holdout_samples = generate_attack_data.draw_real_data(
        dm=dm_holdout, n_samples=n_samples
    )
    outlier_samples = generate_attack_data.draw_real_data(
        dm=dm_train, outliers=True, n_samples=n_samples
    )

    df_train = dm_train.dataset.df[
        dm_train.dataset.df["segment"].isna()
    ].copy()
    outlier_unseen_same_samples = _create_unseen_outliers(
        df_train, dm_train.dataset.feature_mean, mean_factor
    )
    outlier_unseen_diff_samples = _create_unseen_outliers(
        df_train, mean=0, mean_factor=mean_factor
    )

    return MembershipInferenceAttackSamples(
        synthetic_samples=synthetic_samples,
        train_samples=train_samples,
        holdout_samples=holdout_samples,
        outlier_seen_samples=outlier_samples,
        outlier_unseen_same_samples=outlier_unseen_same_samples,
        outlier_unseen_diff_samples=outlier_unseen_diff_samples,
    )


def _create_mia_train_dataset(
    samples: MembershipInferenceAttackSamples,
) -> pd.DataFrame:
    """
    Create training dataset for membership inference attack.
    This dataset is used to train the MIA discriminator

    Args:
        samples (MembershipInferenceAttackSamples): MIA samples.

    Returns:
        pd.DataFrame: Training dataset for MIA.
    """
    df = pd.DataFrame()

    holdout_data = samples.holdout_samples.detach().numpy().tolist()
    synth_data = samples.synthetic_samples.detach().numpy().tolist()

    holdout_target = [0] * len(holdout_data)
    synth_target = [1] * len(synth_data)

    df["tensors"] = holdout_data + synth_data
    df["target"] = holdout_target + synth_target
    df = df.sample(frac=1).reset_index(drop=True)

    return df


def _create_mia_attack_dataset(
    samples: MembershipInferenceAttackSamples,
) -> pd.DataFrame:
    """
    Create attack dataset for membership inference attack.
    This dataset is used to launch the MIA on the seen outliers.


    Args:
        samples (MembershipInferenceAttackSamples): MIA samples.

    Returns:
        pd.DataFrame: Attack dataset for MIA
    """
    df = pd.DataFrame()

    seen_data = samples.outlier_seen_samples.detach().numpy().tolist()
    unseen_same_data = (
        samples.outlier_unseen_same_samples.detach().numpy().tolist()
    )
    unseen_diff_data = (
        samples.outlier_unseen_diff_samples.detach().numpy().tolist()
    )

    seen_target = [1] * len(seen_data)
    unseen_same_target = [0] * len(unseen_same_data)
    unseen_diff_target = [0] * len(unseen_diff_data)

    seen_type = ["seen"] * len(seen_data)
    unseen_same_type = ["unseen_same"] * len(unseen_same_data)
    unseen_diff_type = ["unseen_diff"] * len(unseen_diff_data)

    df["tensors"] = seen_data + unseen_same_data + unseen_diff_data
    df["target"] = seen_target + unseen_same_target + unseen_diff_target
    df["type"] = seen_type + unseen_same_type + unseen_diff_type
    df = df.sample(frac=1).reset_index(drop=True)

    return df


class MembershipInferenceDataset(Dataset):
    def __init__(self, df: pd.DataFrame):
        """
        Membership Inference Dataset

        Args:
            df (pd.DataFrame): Input data
        """
        self.df = df
        self.vector = self.df["tensors"].astype(str).apply(ast.literal_eval)
        self.vector = torch.from_numpy(np.array(self.vector.tolist())).float()
        self.label = self.df["target"].astype(int)

    def __len__(self):
        return len(self.df)

    def __getitem__(self, idx):
        vector = self.vector[idx]
        label = self.label[idx]
        return vector, label


class MembershipInferenceDataModule(pl.LightningDataModule):

    def __init__(
        self,
        model: FaradayModel | PLDiffusion1D,
        dm_train: LCLDataModule,
        dm_holdout: LCLDataModule,
        batch_size: int,
        n_samples: int = 20000,
        mean_factor: int = 20,
        synthetic_samples: torch.Tensor = None,
    ):
        """
        Membership Inference Attack Data Module

        Args:
            model (FaradayModel): Model to generate synthetic samples.
            dm_train (LCLDataModule): Train data module with outliers injected
            dm_holdout (LCLDataModule): Holdout data module
            batch_size (int): Batch size for MIA.
            n_samples (int): Number of training/ holdout samples to train MIA
            mean_factor (int): Mean factor for outliers. Defaults to 20.
        """

        super().__init__()
        self.model = model
        self.dm_train = dm_train
        self.dm_holdout = dm_holdout
        if isinstance(model, FaradayModel):
            self.samples = _create_attack_samples(
                model=model,
                dm_train=dm_train,
                dm_holdout=dm_holdout,
                mean_factor=mean_factor,
                n_samples=n_samples,
            )
        elif isinstance(model, PLDiffusion1D):
            if synthetic_samples is None:
                raise ValueError("Synthetic samples must be provided.")
            self.samples = _create_attack_samples_no_model(
                synthetic_samples=synthetic_samples,
                dm_train=dm_train,
                dm_holdout=dm_holdout,
                mean_factor=mean_factor,
                n_samples=n_samples,
            )
        else:
            raise ValueError("Model not supported")
        self.batch_size = batch_size

    def prepare_data(self):
        pass

    def setup(self, stage=""):

        # Create training and attack datasets
        df_train = _create_mia_train_dataset(self.samples)
        df_attack = _create_mia_attack_dataset(self.samples)

        # Split df_train into train/ eval to prevent overfitting
        df_train_train, df_train_eval = train_test_split(
            df_train, train_size=0.8, random_state=100
        )

        # Reshuffling data
        df_train_train = df_train_train.sample(frac=1, random_state=100)
        df_train_eval = df_train_eval.sample(frac=1, random_state=100)
        df_attack = df_attack.sample(frac=1, random_state=100)

        self.df_train = df_train_train.reset_index(drop=True)
        self.df_eval = df_train_eval.reset_index(drop=True)
        self.df_attack = df_attack.reset_index(drop=True)

        # Create datasets which will be used to instantiate data loaders
        self.train_dataset = MembershipInferenceDataset(df=self.df_train)
        self.eval_dataset = MembershipInferenceDataset(df=self.df_eval)
        self.attack_dataset = MembershipInferenceDataset(df=self.df_attack)

    def train_dataloader(self):
        return DataLoader(
            self.train_dataset, self.batch_size, drop_last=True, shuffle=True
        )

    def eval_dataloader(self):
        return DataLoader(
            self.eval_dataset, self.batch_size, drop_last=True, shuffle=False
        )

    def attack_dataloader(self):
        # Overwrite batch size to full size of attack dataset
        batch_size = len(self.attack_dataset.df)
        return DataLoader(
            self.attack_dataset, batch_size, drop_last=True, shuffle=False
        )


class MembershipInferenceDataModuleDF(pl.LightningDataModule):
    def __init__(self):
        raise NotImplementedError("This class is not implemented yet.")


class MembershipInferenceModule(nn.Module):
    def __init__(self, input_size):
        super().__init__()
        self.input_size = input_size
        self.layers = nn.Sequential(
            nn.Linear(self.input_size, 32),
            nn.Dropout(0.5),
            nn.GELU(),
            nn.Linear(32, 16),
            nn.GELU(),
            nn.Linear(16, 4),
            nn.GELU(),
            nn.Linear(4, 1),
            # nn.Sigmoid(), # using BCELosswithlogits - no sigmoid needed
        )

    def forward(self, x):
        outputs = self.layers(x)
        return outputs


class MembershipInferenceModel(pl.LightningModule):
    def __init__(self, learning_rate, input_size):
        super().__init__()
        self.learning_rate = learning_rate
        self.loss_fn = nn.BCEWithLogitsLoss()
        self.model = MembershipInferenceModule(input_size)
        self.save_hyperparameters()

    def configure_optimizers(self):
        optim = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optim

    def forward(self, kwh):
        model_inputs = kwh
        return self.model(model_inputs)

    def predict(self, kwh, return_raw_prob=False):
        logits = self.forward(kwh)
        probs = torch.sigmoid(logits)

        if return_raw_prob:
            return probs
        else:
            return probs.round()

    def calculate_accuracy(self, pred, label):
        return torchmetrics.functional.classification.binary_accuracy(
            pred, label
        )

    def training_step(self, batch_data):
        tensors = batch_data[0]
        label = batch_data[1]

        logit = self.forward(tensors)
        pred = self.predict(tensors)

        loss = self.loss_fn(logit.squeeze(), label.squeeze().float())
        accuracy = self.calculate_accuracy(pred.squeeze(), label.squeeze())

        self.log(
            "Train BCE Loss", loss, on_epoch=True, on_step=False, prog_bar=True
        )
        self.log(
            "Train Accuracy",
            accuracy,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
        )
        return loss

    def validation_step(self, batch_data, batch_idx):
        tensors = batch_data[0]
        label = batch_data[1]

        logit = self.forward(tensors)
        pred = self.predict(tensors)

        loss = self.loss_fn(logit.squeeze(), label.squeeze().float())
        accuracy = self.calculate_accuracy(pred.squeeze(), label.squeeze())

        self.log(
            "Eval BCE Loss", loss, on_epoch=True, on_step=False, prog_bar=True
        )
        self.log(
            "Eval Accuracy",
            accuracy,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
        )
        return loss


def get_mia_predictions(
    model: MembershipInferenceModel, attack_dm: MembershipInferenceDataModule
):
    model.eval()
    dl = attack_dm.attack_dataloader()
    attack_kwh, attack_label = next(iter(dl))

    # Get predictions - probability and label
    pred_proba = model.predict(attack_kwh, return_raw_prob=True)
    pred_label = model.predict(attack_kwh, return_raw_prob=False)

    # Safe results to dataframe in corresponding records
    df_results = attack_dm.attack_dataset.df.copy()
    df_results["pred_proba"] = pred_proba.squeeze().detach().numpy()
    df_results["pred_label"] = pred_label.squeeze().detach().numpy()
    df_results["true_label"] = attack_label.detach().numpy()

    # Rank predictions
    df_results = df_results.sort_values("pred_proba", ascending=False)
    ones = df_results["true_label"].sum() * [1]
    zeros = (len(df_results) - len(ones)) * [0]
    df_results["rank_pred"] = ones + zeros
    df_results["pred_label"] = df_results["pred_label"].astype(int)
    df_results["true_label"] = df_results["true_label"].astype(int)
    df_results["rank_pred"] = df_results["rank_pred"].astype(int)

    return df_results


def print_mia_results(df: pd.DataFrame):
    df["accuracy"] = df["true_label"] == df["rank_pred"]

    precision = precision_score(df["true_label"], df["rank_pred"])
    f1 = f1_score(df["true_label"], df["rank_pred"])
    acc = accuracy_score(df["true_label"], df["rank_pred"])
    true_positive = len(df.query("true_label==1 and rank_pred==1"))
    false_positive = len(df.query("true_label==0 and rank_pred==1"))
    total_n = len(df)
    total_positive = df["true_label"].sum()

    interval = 1.96 * np.sqrt((precision * (1 - precision)) / total_n)

    print(
        f"""Segment: Precision: {precision:.2f} +/- {interval:.2f},"""
        f"""\nF1: {f1:.2f}, Acc: {acc:.2f},"""
        f"""\nTrue Positives: {true_positive},"""
        f"""\nFalse Positive: {false_positive},"""
        f"""\nTotal N: {total_n},"""
        f"""\nTotal Positive: {total_positive},"""
        f"""\nRandom Precision : {total_positive / total_n:.2f},"""
    )
